import xmlrpclib, socket, errno, logging
import prefs

from gobject import *
from time import sleep
from os import remove
from util import Thread, OrderedDict, logException
from kiwigobject import PropertyObject, gsignal, gproperty, type_register

class __Backend(Thread):
    gproperty("updating", bool, False)
    gproperty("paused", bool, False)
    gproperty("connected", bool, False)
    
    gsignal("updated")
    gsignal("unexpectedDisconnect")
    
    def __init__(self):
        # Initialize threading
        Thread.__init__(self)
        
        # The place where all Download instances are stored
        self.downloads = QueueList()
        
        # Cache information about queued downloads and downloads that are being processed
        # This is used to speed up the updateState method
        self.cache = {
            "queued"        : {},
            "processing"    : {},
        }
        
        # Initialize the remaining properties
        self.speed = 0
        self.eta = 0
        self.remaining = 0
        self.progress = 0
        self.logEntries = []
    
    def start(self):
        logging.debug(_("Starting backend thread..."))
        
        Thread.start(self)
    
    def run(self):
        while not self.threadStopped.isSet():
            if self.connected:
                sleepTime = prefs.lotta["sleep_time"]
                self.updateState()
            else:
                sleepTime = 0.1
            
            sleep(sleepTime)
        
        self.emit("completed")
    
    def stop(self):
        logging.debug(_("Shutting down backend thread..."))
        
        self.disconnectFromHella()
        Thread.stop(self)
    
    def _getServerProxy(self, address, port, password, silent = False):
        location = address + ":" + str(port)
        uri = "http://hellanzb:" + password + "@" + location + "/"
        
        try:
            proxy = xmlrpclib.ServerProxy(uri)
            list = proxy.system.listMethods()
            
            for cmd in ["cancel", "clear", "continue", "clear", "continue", "dequeue",
                "down", "enqueue", "enqueuenewzbin", "enqueueurl", "force", "maxrate",
                "move", "pause", "process", "setrarpass", "shutdown", "status", "up"]:
                assert cmd in list
        
        except xmlrpclib.ProtocolError, e:
            if e.errcode == 401:
                raise WrongPasswordError, _("The specified password is invalid. Please check your input.")
            else:
                raise ConnectionError, str(e.errcode) + ": " + e.errmsg
        
        except (socket.gaierror, socket.error), e:
            if not silent:
                logging.debug(str(e[0]) + ": " + e[1])
            
            if e[0] == errno.ECONNREFUSED:
                raise NotRunningError, _("The HellaNZB daemon doesn't seem to be running at %s.") % (location)
            else:
                raise ConnectionError, _("Could not connect to %s. This might be due to network issues.") % (address)
        
        except AssertionError, e:
            raise UnsupportedDaemonError, _("You don't seem to be running a compatible HellaNZB daemon. Plase make sure you have installed the latest version.")
        
        else:
            setattr(proxy, "resume", getattr(proxy, "continue"))
            return proxy
    
    def connectToHella(self, address, port, password):
        location = address + ":" + str(port)
        
        try:
            self.setProxy(self._getServerProxy(address, port, password))
        except:
            logging.error(_("Could not connect to the HellaNZB daemon at %s.") % (location))
            raise
        else:
            self.connected = True
            logging.info(_("Connected to the HellaNZB daemon at %s.") % (location))
    
    def setProxy(self, proxy):
        self.__server = proxy
    
    def disconnectFromHella(self):
        self.__server = None
        
        self.connected = False
    
    def shutdown(self):
        self.__server.shutdown()
    
    def checkConnection(self, address, port, password):
        return self._getServerProxy(address, port, password, True)
    
    def enqueue(self, name):
        try:
            self.updateState(self.__server.enqueue(name))
            remove(name)
        except:
            logging.error(_("Could not add file to the download queue."))
        else:
            logging.info(_("Added %s to the download queue.") % (name))
    
    def newzbinEnqueue(self, id):
        try:
            self.updateState(self.__server.enqueuenewzbin(id))
        except:
            logging.error(_("Could not add file to the download queue."))
        else:
            logging.info(_("Added %s to the download queue.") % (id))
    
    def dequeue(self, download):
        try:
            if download:
                if download.state == Download.STATE_DOWNLOADING:
                    self.cancel()
                else:
                    # dequeue(id) doesn't return the whole state information
                    self.__server.dequeue(download.id)
                    self.updateState()
                
                logging.info(_("Removed '%s' from queue.") % (download))
        except:
            logging.error(_("Could not dequeue item."))
    
    def setRarPass(self, id, password):
        """This methods sets the password for the rar archive(s) in the item with the id provided."""
        self.updateState(self.__server.setrarpass(id, password))
    
    def cancel(self):
        try:
            state = self.__server.cancel()
            
            # Make sure the canceled download isn't marked as finished.
            # Todo: Could be implemented more elegantly
            self.downloads.remove(self.downloads.getActiveDownload())
            
            self.updateState(state)
        except:
            logging.error(_("Could not cancel the current download."))
        else:
            logging.info(_("Canceled current download."))
    
    def clear(self, andCancel = False):
        try:
            self.updateState(self.__server.clear(andCancel))
        except:
            logging.error(_("Could not clear the queue."))
        else:
            logging.info(_("Cleared queue."))
    
    def pause(self):
        try:
            self.paused = True
            self.updateState(self.__server.pause())
        except:
            logging.error(_("Could not pause the download."))
        else:
            logging.info(_("Paused downloads."))
    
    def resume(self):
        try:
            self.paused = False
            self.updateState(self.__server.resume())
        except:
            logging.error(_("Could not resume the download."))
        else:
            logging.info(_("Resuming downloads."))
    
    def moveUp(self, download, shift = 1):
        try:
            if download and shift:
                if self.downloads.getQueued().index(download) < shift:
                    self.updateState(self.__server.force(download.id))
                    logging.info(_("Forced '%s' to begin.") % (download))
                else:
                    self.__server.up(download.id, shift)
                    self.updateState()
                    logging.info(_("Moved '%s' up.") % (download))
        except:
            logging.error(_("Could not move item."))
    
    def moveDown(self, download, shift = 1):
        try:
            if download and shift:
                if self.downloads.getActiveDownload() is download:
                    queued = self.downloads.getQueued()
                    
                    if len(queued):
                        firstQueued = queued.first()
                        
                        self.updateState(self.__server.force(firstQueued.id))
                        
                        if shift > 1:
                            self.__server.down(download.id, shift - 1)
                            self.updateState()
                        
                        logging.info(_("Forced '%s' to begin.") % (firstQueued))
                else:
                    self.__server.down(download.id, shift)
                    self.updateState()
                    logging.info(_("Moved '%s' down.") % (download))
        except:
            logging.error(_("Could not move item."))
    
    def moveTo(self, download, target):
        if not download.movable():
            logging.error(_("Can't move finished downloads or downloads that are being processed."))
        
        queue = self.downloads.getIncompleteDownloads()
        downloadIndex = queue.index(download)
        
        if isinstance(target, Download):
            position = queue.index(target)
        elif target == "bottom":
            position = queue.len()
        elif target == "top":
            position = 0
        else:
            position = target
        
        if position > downloadIndex + 1:
            download.moveDown(position - downloadIndex - 1)
        elif position < downloadIndex:
            download.moveUp(downloadIndex - position)
    
    def setMaxRate(self, maxRate = 0):
        try:
            self.__server.maxrate(maxRate)
        except:
            logging.error(_("Could not set the maxium download speed."))
        else:
            if maxRate:
                logging.info(_("Set the maximum download speed to %s KiB/s.") % (maxRate))
            else:
                logging.info(_("Removed download speed limit."))
    
    def getMaxRate(self):
        try:
            return self.__server.maxrate()
        except:
            return 0
    
    def updateState(self, state = None):
        if self.updating:
            return
        
        self.updating = True
        
        try:
            if state is None:
                try:
                    state = self.__server.status()
                except:
                    logging.warning(_("Could not get an update from the HellaNZB daemon, is it (still) running?"))
            
            if not state:
                self.connected = False
                self.emit("unexpectedDisconnect")
                self.updating = False
                return
            
            self.paused = state["is_paused"]
            self.speed = state["rate"]
            self.eta = state["eta"]
            self.remaining = state["queued_mb"]
            self.progress = state["percent_complete"]
            self.logEntries = state["log_entries"]
            
            unfinishedOrder = []
            finishedOrder = self.downloads.getFinished().keys()
            
            def updateDownloadList(list, state):
                updatedDownloads = []
                
                for downloadInfo in list:
                    id = downloadInfo["id"]
                    updatedDownloads.append(id)
                    
                    if id in self.downloads:
                        download = self.downloads[id]
                        download.update(downloadInfo)
                        download.state = state
                    else:
                        self.downloads.addFromHella(downloadInfo, state)
                
                unfinishedOrder.extend(updatedDownloads)
                
                return updatedDownloads
            
            queued = state["queued"]
            downloading = state["currently_downloading"]
            processing = state["currently_processing"]
            
            if not self.cache["processing"] == processing:
                updatedIDs = updateDownloadList(processing, Download.STATE_PROCESSING)
                
                for id in self.downloads.getProcessing().keys():
                    if not id in updatedIDs:
                        self.downloads[id].state = Download.STATE_FINISHED
                        finishedOrder.append(id)
                
                self.cache["processing"] = processing
            else:
                unfinishedOrder.extend(self.downloads.getProcessing().keys())
            
            lastActiveDownload = self.downloads.getActiveDownload()
            
            updateDownloadList(downloading, Download.STATE_DOWNLOADING)
            
            if not self.cache["queued"] == queued:
                updateDownloadList(queued, Download.STATE_QUEUED)
                self.cache["queued"] = queued
            else:
                unfinishedOrder.extend(self.downloads.getQueued().keys())
            
            if lastActiveDownload and not lastActiveDownload.id in unfinishedOrder:
                lastActiveDownload.state = Download.STATE_FINISHED
                finishedOrder.append(lastActiveDownload.id)
            
            finishedOrder.extend(unfinishedOrder)
            self.downloads.reorder(finishedOrder)
            
            self.emit("updated")
        except:
            logging.error(_("Could not update LottaNZB's state."))
            logException()
        
        self.updating = False
    
    def clearFinished(self):
        self.downloads.clearFinished()
        logging.info(_("Cleared the finished items."))
    
    def formatETA(self, eta):
        hours = int(eta / (60 * 60))
        minutes = int((eta - (hours * 60 * 60)) / 60)
        seconds = eta - (hours * 60 * 60) - (minutes * 60)
        
        return "%.2d:%.2d:%.2d" % (hours, minutes, seconds)

class ConnectionError(Exception):
    pass

class WrongPasswordError(ConnectionError):
    pass

class NotRunningError(ConnectionError):
    pass

class UnsupportedDaemonError(ConnectionError):
    pass

__singleton = None
def Backend():
    global __singleton
    
    if __singleton is None:
        __singleton = __Backend()
    
    return __singleton 

class DownloadList(OrderedDict):
    def add(self, download):
        self[download.id] = download
    
    def remove(self, download):
        del self[download.id]

class QueueList(DownloadList):
    def addFromHella(self, info, state):
        download = Download(info, state)
        
        self.add(download)
        
        return download
    
    def _getByState(self, state):
        downloads = DownloadList()
        
        for download in self.values():
            if download.state == state:
                downloads.add(download)
        
        return downloads
    
    def getQueued(self):
        return self._getByState(Download.STATE_QUEUED)
    
    def getActiveDownload(self):
        downloading = self._getByState(Download.STATE_DOWNLOADING)
        
        if downloading:
            return downloading.first()
    
    def getProcessing(self):
        return self._getByState(Download.STATE_PROCESSING)
    
    def getFinished(self):
        return self._getByState(Download.STATE_FINISHED)
    
    # Todo: Find a better name for this method
    def getIncompleteDownloads(self):
        queue = DownloadList()
        activeDownload = self.getActiveDownload()
        
        if activeDownload:
            queue.add(activeDownload)
        
        queue.update(self.getQueued())
        
        return queue
    
    def clearFinished(self):
        map(self.remove, self.getFinished().values())
    
    def getTotalSize(self):
        totalsize = 0
        
        for queued in self.getQueued().values():
            totalsize += queued.remaining
        
        activeDownload = self.getActiveDownload()
        
        if activeDownload:
            totalsize += activeDownload.remaining
        
        return totalsize
    
    def getTotalEta(self):
        activeDownload = self.getActiveDownload()
        
        if activeDownload and activeDownload.remaining:
            return int(Backend().eta * self.getTotalSize() / activeDownload.remaining)
        
        return 0

class Download(GObject):
    id = property(type = int)
    name = property(type = str)
    nzb_name = property(type = str)
    category = property(type = str)
    remaining = property(type = int, minimum = 0)
    progress = property(type = int, minimum = 0, maximum = 100)
    recovery = property(type = bool, default = False)
    password = property(type = str)
    size = property(type = int, minimum = 0)
    state = property(type = int)
    
    def __init__(self, hellaInfo, state):
        GObject.__init__(self)
        
        self.connect("notify::state", self._handleStateChange)
        self.connect("notify::size", self._handleSizeChange)
        
        self.id = hellaInfo["id"]
        self.nzb_name = hellaInfo["nzbName"]
        
        parts = self.nzb_name.split(Download.CATEGORY_SEPARATOR)
        
        if len(parts) > 1:
            self.category = parts[0].strip()
            self.name = parts[1].strip()
        else:
            self.category = _("Unknown")
            self.name = self.nzb_name
        
        self.update(hellaInfo)
        self.state = state
    
    def update(self, hellaInfo):
        self.recovery = hellaInfo["is_par_recovery"]
        
        if "total_mb" in hellaInfo:
            self.size = hellaInfo["total_mb"]
        
        if "rarPassword" in hellaInfo:
            self.password = hellaInfo["password"]
    
    def dequeue(self):
        Backend().dequeue(self)
    
    def moveTo(self, target):
        Backend().moveTo(self, target)
    
    def moveUp(self, shift = 1):
        Backend().moveUp(self, shift)
    
    def moveDown(self, shift = 1):
        Backend().moveDown(self, shift)
    
    def _handleStateChange(self, *args):
        if self.state == Download.STATE_PROCESSING or self.state == Download.STATE_FINISHED:
            self.progress = 100
            self.remaining = 0
        elif self.state == Download.STATE_DOWNLOADING:
            self.progress = Backend().progress
            self.remaining = Backend().remaining
    
    def _handleSizeChange(self, *args):
        if self.size != 0 and not self.progress and not self.remaining:
            self.remaining = self.size
    
    def movable(self):
        return not (self.state == Download.STATE_FINISHED or self.state == Download.STATE_PROCESSING)
    
    def __str__(self):
        return self.name
    
    STATE_QUEUED = 0
    STATE_DOWNLOADING = 1
    STATE_PROCESSING = 2
    STATE_FINISHED = 3
    
    CATEGORY_SEPARATOR = " -- "
    CATEGORIES = [_("Music"), _("DVD Movie"), _("XVID Movie"), _("TV Series"), _("HD Movie"), _("Game"), _("Software"), _("Other")]
