import sys
import gtk, os, logging
from shutil import move
import log, util, paths, backend, prefs, prefsgui, modes
from backend import Backend, Download
import warnings

try:
    import gnome
except:
    GNOME_MODULE_AVAILABLE = False
else:
    GNOME_MODULE_AVAILABLE = True

class MainWindow(util.Window):
    def __init__(self):
        if GNOME_MODULE_AVAILABLE:
            gnome.program_init("lottanzb", "0.2.1-pre", properties = { gnome.PARAM_APP_DATADIR : "/usr/share" })
        
        # Ignore GtkWarnings
        warnings.simplefilter("ignore", gtk.Warning)
        
        # Now we'll add a notification area icon (statusIcon)
        self.statusIcon = gtk.StatusIcon()
        self.statusIcon.set_from_icon_name("lottanzb")
        self.statusIcon.set_tooltip(_("LottaNZB Client"))
        
        # Try to enable RGBA support
        # The method gtk.gdk.Drawable#get_rgba_colormap is available in PyGTK 2.10 and above
        try:
            gtk_screen = self.statusIcon.get_screen()
            colormap = gtk_screen.get_rgba_colormap()
            
            if colormap is None:
                colormap = gtk_screen.get_rgb_colormap()
            
            gtk.widget_set_default_colormap(colormap)
        except:
            pass
        
        util.Window.__init__(self, "MainWindow", {
            "on_MainMenuClose_activate"             : self.toggle,
            "on_MainWindow_delete_event"            : self.toggle,
            "on_MainMenuEditPreferences_activate"   : self.showPreferences,
            "on_MainMenuEditReconfigure_activate"   : self.reconfigureClient,
            "on_MainMenuFileAdd_activate"           : self.showAddFileDialog,
            "on_MainMenuFileQuit_activate"          : self.destroy,
            "on_MainMenuContents_activate"          : self.showHelpContent,
            "on_MainMenuAbout_activate"             : self.showAboutDialog,
            "on_MainMenuBug_activate"               : self.reportBug,
            "on_MainMenuLog_activate"               : self.showLogWindow,
            "on_MainToolbarAdd_clicked"             : self.showAddFileDialog,
            "on_MainToolbarRemove_clicked"          : self.delFile,
            "on_MainToolbarUp_clicked"              : self.moveUp,
            "on_MainToolbarDown_clicked"            : self.moveDown,
            "on_MainToolbarPause_toggled"           : self.togglePause,
            "on_MainToolbarClear_clicked"           : self.clearFinished,
            "on_MainToolbarOpenDownloadDir_clicked" : self.openDownloadDirectory,
            "on_MainQueueTree_drag_data_get"        : self.dragGet,
            "on_MainQueueTree_drag_data_received"   : self.dragReceived, 
            "on_MainQueueTree_cursor_changed"       : self.updateToolbarButtons,
            "on_MainQueueTree_button_press_event"   : self.showRightClickMenu,
            "on_StatusMenuAdd_activate"             : self.showAddFileDialog,
            "on_StatusMenuQuit_activate"            : self.destroy,
            "on_StatusMenuOpen_toggled"             : self.toggle,
            "on_MainMenuFileNewzbinAdd_activate"    : self.showAddNewzbinDialog
        })
        
        self.window.resize(prefs.lotta["window_width"] or 700, prefs.lotta["window_height"] or 300)
        
        self.info = self.get_widget_dict(["Remaining", "TotalRemaining", "TotalRemainingIcon", "Speed"], "MainInfo")
        self.toolbar = self.get_widget_dict(["Up", "Down", "Pause", "Remove", "OpenDownloadDir"], "MainToolbar")
        self.menu = self.get_widget_dict(["FileNewzbinAdd", "EditPreferences"], "MainMenu")
        
        rightClickMenuHelper = util.WidgetHelper("MainRightClick", {
            "on_MainRightClickRemove_activate"      : self.delFile,
            "on_MainRightClickUp_activate"          : self.moveUp,
            "on_MainRightClickDown_activate"        : self.moveDown
        })
        
        self.rightClick = rightClickMenuHelper.get_widget_dict(["Up", "Down", "Remove", "EditName", "SetRarPassword"], "MainRightClick")
        self.mainRightClick = rightClickMenuHelper.widget
        
        upMenu = util.WidgetHelper("MainToolbarUpMenu", {
            "on_MainToolbarToTop_activate"          : self.moveToTop,
            "on_MainToolbarMoveUp_activate"         : self.moveUp
        }).widget
        
        self.toolbar["up"].set_menu(upMenu)
        
        for button in ["up", "down", "remove"]:
            self.toolbar[button].set_sensitive(False)
        
        self.queueTree = self.get_widget("MainQueueTree")
        
        self.queueList = QueueListModel()
        self.queueTree.set_model(self.queueList)
        
        rowTarget = ("DOWNLOAD_ROW", gtk.TARGET_SAME_WIDGET, 0)
        
        self.queueTree.enable_model_drag_source(gtk.gdk.BUTTON1_MASK, [rowTarget], gtk.gdk.ACTION_DEFAULT | gtk.gdk.ACTION_MOVE) 
        self.queueTree.enable_model_drag_dest([rowTarget, ("text/plain", 0, 1)], gtk.gdk.ACTION_DEFAULT | gtk.gdk.ACTION_MOVE)
        
        self.addQueueColumn(_("Name"), 1)
        self.addQueueColumn(_("Category"), 2)
        self.addQueueColumn(_("Size"), 3)
        self.addQueueColumnProgressBar(_("Progress"), 4, 5)
        
        statusMenuHelper = util.WidgetHelper("StatusMenu", {
            "on_StatusMenuAdd_activate"             : self.showAddFileDialog,
            "on_StatusMenuQuit_activate"            : self.destroy,
            "on_StatusMenuOpen_toggled"             : self.toggle
        })
        
        self.statusMenu = statusMenuHelper.widget
        self.statusMenuOpen = statusMenuHelper.get_widget("StatusMenuOpen")
        
        # Now we connect the menu to the statusIcon
        self.statusIcon.connect("popup-menu", self.showMenu, self.statusMenu)
        self.statusIcon.connect("activate", self.toggle)
        
        if prefs.lotta["start_minimized"]:
            self.hide()
            self.toggling = True
            self.statusMenuOpen.set_active(False)
            self.toggling = False
        else:
            self.show()
        
        Backend().connect("updated", self.update)
        
        if modes.getActiveMode():
            self.handleModeChanged()
        
        modes.events.connect("mode-changed", self.handleModeChanged)
    
    def handleModeChanged(self, *args):
        mode = modes.getActiveMode()
        prefsLoaded = bool(mode and mode.hellaPrefs.loaded)
        
        try:
            assert prefsLoaded
            
            global gnomevfs
            import gnomevfs
        except:
            self.toolbar["opendownloaddir"].hide()
        else:
            self.toolbar["opendownloaddir"].show()
        
        activateNewzbinSupport = bool(mode and (modes.RemoteFrontendMode.isActive() or mode.hellaPrefs.hasNewzBinSupport()))
        self.menu["filenewzbinadd"].set_sensitive(activateNewzbinSupport)
        
        self.menu["editpreferences"].set_sensitive(modes.StandaloneMode.isActive())
    
    def openDownloadDirectory(self, widget):
        try:
            gnomevfs.url_show("file://" + prefs.hella.DEST_DIR)
        except gnomevfs.NotFoundError:
            util.Message.error(_("Error"), _("Could not find the download directory."))
        except:
            pass
    
    def showRightClickMenu(self, button, event):
        """This method shows the right click menu, but only if the third button (that's right indeed, the right mouse button) is clicked."""
        if event.button == 3:
            self.mainRightClick.popup(None, None, None, event.button, event.time)
    
    def dragGet(self, treeview, context, selection, info, timestamp):
        model, selection_iter = self.queueTree.get_selection().get_selected()
        path = model.get_path(selection_iter)
        
        if model.row_draggable(path):
            return model.drag_data_get(path, selection)
        
        return
    
    def dragReceived(self, treeview, context, x, y, selection, info, timestamp):
        drop_info = treeview.get_dest_row_at_pos(x, y)
        model = treeview.get_model()
        path = None
        
        if not selection.data:
            return
        
        if drop_info:
            path, position = drop_info
            
            if position == gtk.TREE_VIEW_DROP_AFTER or position == gtk.TREE_VIEW_DROP_INTO_OR_AFTER:
                path = (path[0] + 1, )
        
        if info == 0:
            if model.row_drop_possible(path, selection):
                model.drag_data_received(path, selection)
        elif info == 1:
            self.handleNZBDrop(selection.data)
        
        if context.action == gtk.gdk.ACTION_MOVE:
            context.finish(True, True, timestamp)
        
        return
    
    def showAddNewzbinDialog(self, *args):
        if not hasattr(self, "addNewzbinDialog") or not self.addNewzbinDialog.widget.has_user_ref_count:
            self.addNewzbinDialog = AddNewzbinDialog()
    
    def toggle(self, *args):
        if util.Window.toggle(self):
            self.toggling = True
            self.statusMenuOpen.set_active(self.visible())
            self.toggling = False
            return True
        
        return False
    
    def update(self, *args):
        backend = Backend()
        activeDownload = backend.downloads.getActiveDownload()
        
        if activeDownload:
            self.toolbar["pause"].set_active(backend.paused)
            self.info["speed"].set_text(_("%s KB/s") % (backend.speed))
            self.info["remaining"].set_text(_("Remaining: %s MB") % (str(activeDownload.remaining)) +  " - " + str(backend.formatETA(backend.eta)))
            
            if backend.downloads.getQueued():
                self.info["totalremaining"].show()
                self.info["totalremainingicon"].show()
                self.info["totalremaining"].set_text(_("Total remaining: %s MB") % (str(backend.downloads.getTotalSize())) + " - " + str(backend.formatETA(backend.downloads.getTotalEta())))
            else:
                self.info["totalremaining"].hide()
                self.info["totalremainingicon"].hide()
        else:
            self.info["speed"].set_text(_("None"))
            self.info["remaining"].set_markup(_("None"))
        
        selectedDownload = self.getSelectedDownload()
        
        self.queueList = QueueListModel()
        self.queueTree.set_model(self.queueList)
        
        if selectedDownload:
            selection_iter = self.queueList.get_iter_from_download(selectedDownload)
            
            if selection_iter:
                self.queueTree.get_selection().select_iter(selection_iter)
        
        self.updateToolbarButtons()
    
    def getSelectedDownload(self):
        model, selection_iter = self.queueTree.get_selection().get_selected()
        
        if selection_iter:
            id = model.get_value(selection_iter, 0)
            
            if id in Backend().downloads:
                return Backend().downloads[id]
    
    def updateToolbarButtons(self, *args):
        download = self.getSelectedDownload()
        active = download is not None and download.movable()
        
        actionStates = {}
        
        for button in ["up", "down", "remove"]:
            actionStates[button] = active
        
        if download:
            if download.state == Download.STATE_DOWNLOADING:
                actionStates["up"] = False
            if Backend().downloads.last() is download:
                actionStates["down"] = False
        
        for action, actionState in actionStates.iteritems():
            self.toolbar[action].set_sensitive(actionState)
            self.rightClick[action].set_sensitive(actionState)
    
    def showMenu(self, status_icon, button, activate_time, menu):
        self.statusMenu.popup(None, None, gtk.status_icon_position_menu, button, activate_time, status_icon)
    
    def addQueueColumn(self, title, columnID):
        column = gtk.TreeViewColumn(title, gtk.CellRendererText(), text=columnID)
        column.set_resizable(True)
        column.set_expand(True)
        self.queueTree.append_column(column)
    
    def addQueueColumnProgressBar(self, title, columnID, progress):
        column = gtk.TreeViewColumn(title, gtk.CellRendererProgress(), text=columnID, value=progress)
        column.set_resizable(False)
        column.set_expand(False)
        column.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
        column.set_fixed_width(200)
        self.queueTree.append_column(column)
    
    def destroy(self, *args):
        size = self.window.get_size()
        
        prefs.lotta.update({
            "window_width"  : size[0],
            "window_height" : size[1]
        })
        
        self.hide()
        sys.exit()
    
    def moveDown(self, void):
        download = self.getSelectedDownload()
        
        if download:
            download.moveDown()
    
    def moveUp(self, void):
        download = self.getSelectedDownload()
        
        if download:
            download.moveUp()
    
    def moveToTop(self, void):
        download = self.getSelectedDownload()
        
        if download:
            download.moveTo("top")
    
    def clearFinished(self, void):
        Backend().clearFinished()
    
    def togglePause(self, void):
        if self.toolbar["pause"].get_active():
            Backend().pause()
        else:
            Backend().resume()
    
    def delFile(self, void):
        download = self.getSelectedDownload()
        
        if download and util.Message.question(_("Are you sure you want to cancel this download?")):
            download.dequeue()
    
    def showAddFileDialog(self, widget, file = None):
        if not hasattr(self, "addFileDialog") or not self.addFileDialog.widget.has_user_ref_count:
            self.addFileDialog = AddFileDialog(file)
    
    def handleNZBDrop(self, data):
        data = data.replace("file://", "")
        # data = data.replace("\r\n", "")
        
        data = data.strip()
        
        if len(data.split()) == 1:
            AddFileDialog(data.replace("%20", " ").strip())
        else:
            for file in data.split():
                if file.lower().endswith(".nzb"):
                    Backend().enqueue(file.replace("%20", " ").strip())
    
    def showHelpContent(self, *args):
        def showOnlineDocumentation():
            import webbrowser
            
            webbrowser.open_new_tab("http://lottanzb.org/index.php?option=com_content&view=article&id=9&Itemid=7")
        
        def getHelpLocale():
            try:
                import locale
                
                currentLocale = locale.getlocale()[0]
                
                if currentLocale == "C":
                    return currentLocale
                else:
                    helpLocale = currentLocale[:2]
                    docFile = paths.docsDir(helpLocale + "/lottanzb.xml")
                    
                    if os.path.isfile(docFile):
                        return helpLocale
                    else:
                        return "C"
            except:
                return "C"
        
        def showHelpUsingPopen():
            try:
                os.popen4("yelp " + paths.docsDir(getHelpLocale() + "/lottanzb.xml"))
            except:
                showOnlineDocumentation()
        
        if paths.isInstalled():
            if GNOME_MODULE_AVAILABLE:
                try:
                    gnome.help_display("lottanzb")
                except:
                    showOnlineDocumentation()
            else:
                showHelpUsingPopen()
        else:
            showHelpUsingPopen()
    
    def showAboutDialog(self, *args):
        if not hasattr(self, "aboutDialog") or not self.aboutDialog.widget.has_user_ref_count:
            self.aboutDialog = AboutDialog()
    
    def reportBug(self, *args):
        import webbrowser
        
        webbrowser.open_new_tab("http://bugs.launchpad.net/lottanzb/+filebug")
    
    def showPreferences(self, *args):
        if modes.StandaloneMode().isActive():
            prefsgui.StandaloneModeDialog()
    
    def reconfigureClient(self, *args):
        modes.showSelectionWindow()
    
    def showLogWindow(self, *args):
        if not hasattr(self, "logWindow") or not self.logWindow.widget.has_user_ref_count:
            self.logWindow = log.LogWindow()

class QueueListModel(gtk.ListStore):
    def __init__(self):
        gtk.ListStore.__init__(self, int, str, str, str, str, int)
        
        self.fill()
    
    def fill(self):
        self.clear()
        
        for id, download in Backend().downloads.items():
            if download.size:
                size = _("%s MB") % (str(download.size))
            else:
                size = ""
            
            if download.state == Download.STATE_FINISHED:
                state = _("Finished")
            elif download.state == Download.STATE_PROCESSING:
                if download.recovery:
                    state = _("Repairing and extracting")
                else:
                    state = _("Processing")
            elif download.state == Download.STATE_DOWNLOADING:
                state = "%s %%" % (str(download.progress))
                
                if Backend().paused:
                    state = _("Paused at: %s%%") % (download.progress)
                elif download.recovery:
                    state = _("Downloading PAR recovery files: %s") % (download.progress)
            else:
                if download.recovery:
                    state = _("Queued: Need PAR recovery")
                else:
                    state = _("Queued")
            
            self.append([download.id, download.name, download.category, size, state, download.progress])
    
    def get_iter_from_download(self, download):
        for path, row in enumerate(self):
            if row[0] == download.id:
                return self.get_iter(path)
    
    def get_download_from_path(self, path):
        return Backend().downloads.getByIndex(path[0])
    
    def row_draggable(self, path):
        return self.get_download_from_path(path).movable()
    
    def drag_data_delete(self, path):
        return True
    
    def drag_data_get(self, path, selection):
        selection.set(selection.target, 8, str(self.get_download_from_path(path).id))
        
        return True
    
    def drag_data_received(self, dest, selection):
        if dest:
            # Just in case the destination isn't valid
            try:
                targetDownload = self.get_download_from_path(dest)
                
                if targetDownload.movable():
                    target = targetDownload
                else:
                    target = "top"
            except:
                target = "bottom"
        else:
            target = "bottom"
        
        Backend().downloads[int(selection.data)].moveTo(target)
        Backend().updateState()
        
        return True
    
    def row_drop_possible(self, dest, selection):
        return True

class AboutDialog(util.Window):
    def __init__(self):
        util.Window.__init__(self, "AboutDialog", {
            "on_AboutDialog_response" : self.handleResponse
        })
        
        import __init__
        self.window.set_version(__init__.__version__)
        
        self.show()
    
    def handleResponse(self, widget, response):
        if response == gtk.RESPONSE_CANCEL:
            self.destroy()

class AddFileDialog(util.Window):
    def __init__(self, file = None):
        util.Window.__init__(self, "AddFileDialog", {
            "on_AddFileDialog_response"             : self.handleResponse,
            "on_AddFileDialog_selection_changed"    : self.handleSelectionChange
        })
        
        self.fields = {
            "name"      : self.get_widget("AddFileDialogFileName"),
            "category"  : self.getCategoryBox()
        }
        
        filter = gtk.FileFilter()
        filter.add_pattern("*.nzb")
        filter.add_pattern("*.NZB")
        
        self.window.set_filter(filter)
        
        if not file is None:
            try:
                self.window.select_filename(file)
            except:
                logging.error(_("Could not set the filename."))
                return
        
        self.show()
    
    def handleResponse(self, widget, response):
        if response == gtk.RESPONSE_OK:
            self.enqueueSelection()
        else:
            self.destroy()
    
    def handleSelectionChange(self, *args):
        names = self.window.get_filenames()
        
        if len(names) == 1:
            if names[0][-4:].lower() == ".nzb":
                name = names[0][:-4].split("/")[-1].replace(".", " ").replace("_", " ")
                self.fields["name"].set_text(name)
                self.fields["name"].set_sensitive(True)
        elif len(names) >= 1:
            self.fields["name"].set_text(_("Selected multiple files"))
            self.fields["name"].set_sensitive(False)
        else:
            self.fields["name"].set_text("")
    
    def enqueueSelection(self):
        files = self.window.get_filenames()
        
        if files:
            for file in files:
                if not os.path.isdir(file):
                    category = self.fields["category"].get_active_text()
                    
                    if len(files) == 1:
                        name = self.fields["name"].get_text()
                    else:
                        name = file[:-4].split("/")[-1]
                    
                    move(file, paths.tempDir() + category + Download.CATEGORY_SEPARATOR + name + ".nzb")
                    returnvalue = paths.tempDir() + category + Download.CATEGORY_SEPARATOR + name + ".nzb"
                    Backend().enqueue(returnvalue)
                
                else:
                    self.window.set_current_folder(self.window.get_filename())
            
            self.destroy()
    
    def getCategoryBox(self):
        categoryBox = gtk.combo_box_new_text()
        
        for category in Download.CATEGORIES:
            categoryBox.append_text(category)
        
        categoryBox.set_active(7)
        categoryBox.show()
        
        self.get_widget("AddFileDialogTable").attach(categoryBox, 1, 2, 1, 2)
        
        return categoryBox

class AddNewzbinDialog(util.Window):
    def __init__(self):
        util.Window.__init__(self, "AddNewzbinDialog", {
            "on_AddNewzbinDialog_response"       : self.handleResponse
        })
        
        self.fields = self.get_widget_dict(["ID"])
        self.buttons = self.get_widget_dict(["Add", "Cancel"])
        
        import re
        
        def handleValidationChange(form, *args):
            self.buttons["add"].set_sensitive(form.valid)
        
        form = util.LottaForm(self.fields, { "id": util.LottaFormCheckers["numerical"] })
        form.connect("notify::valid", handleValidationChange)
        
        form.validate()
        
        self.show()
    
    def handleResponse(self, widget, response):
        if response == gtk.RESPONSE_OK:
            self.enqueueSelection()
        else:
            self.destroy()
    
    def enqueueSelection(self):
        id = int(self.fields["id"].get_text())
        
        if id:
            Backend().newzbinEnqueue(id)
            self.destroy()
