import logging, gtk

from util import Window
from time import strftime, localtime
from gobject import GObject
from backend import Backend
from kiwigobject import PropertyObject, gsignal, gproperty, type_register

RECORD_SOURCE_LOTTA = 0
RECORD_SOURCE_HELLA = 1

class LogStore(PropertyObject, GObject):
    __gtype_name__ = "LogStore"
    
    def __init__(self):
        self.records = []
        
        GObject.__init__(self)
        PropertyObject.__init__(self)
    
    def filterByLevel(self, level):
        store = LogStore()
        
        for record in self.records:
            if record.levelno >= level:
                store.records.append(record)
        
        return store
    
    def filterBySource(self, source):
        store = LogStore()
        
        for record in self.records:
            if record.source == source:
                store.records.append(record)
        
        return store
    
    def getLottaStore(self):
        return self.filterBySource(RECORD_SOURCE_LOTTA)
    
    def getHellaStore(self):
        return self.filterBySource(RECORD_SOURCE_HELLA)

type_register(LogStore)

class __Log(LogStore):
    gsignal("changed")
    gproperty("lastHellaRecordIndex", int, default = -1, minimum = -1)
    
    def __init__(self):
        LogStore.__init__(self)
        
        Backend().connect("updated", self.parseHellaLog)
        
        def handleConnectedChange(*args):
            if Backend().connected == False:
                self.clearHellaLog()
        
        Backend().connect("notify::connected", handleConnectedChange)
    
    def parseHellaLog(self, *args):
        entries = Backend().logEntries
        currentIndex = len(entries) - 1
        
        if self.lastHellaRecordIndex > currentIndex:
            self.clearHellaLog()
        elif self.lastHellaRecordIndex < currentIndex:
            for i in range(self.lastHellaRecordIndex + 1, currentIndex + 1):
                entry = entries[i]
                
                for level, message in entry.iteritems():
                    try:
                        levelno = getattr(logging, level)
                    except:
                        levelno = logging.INFO
                    
                    self.records.append(makeLogRecord(RECORD_SOURCE_HELLA, {
                        "msg" : message,
                        "levelno" : levelno
                    }))
            
            self.emit("changed")
        
        self.lastHellaRecordIndex = currentIndex
    
    def clear(self):
        self.records[:] = []
        self.emit("changed")
    
    def clearHellaLog(self):
        lottaRecords = self.getLottaStore().records
        
        self.clear()
        self.records.extend(lottaRecords)
        
        self.lastHellaRecordIndex = -1
        self.emit("changed")

log = __Log()

class LottaHandler(logging.Handler):
    def emit(self, record):
        setLogRecordSource(record, RECORD_SOURCE_LOTTA)
        log.records.append(record)
        log.emit("changed")

def makeLogRecord(source, dict):
    record = logging.LogRecord(None, None, "", 0, "", (), None)
    record.__dict__.update(dict)
    setLogRecordSource(record, source)
    
    return record

def setLogRecordSource(record, source):
    record.__dict__["source"] = source

class LogWindow(Window):
    def __init__(self):
        Window.__init__(self, "LogWindow", {
            "on_LogWindow_delete_event"                 : self.destroy,
            "on_LogWindowToolbarButtonSave_clicked"     : self.save,
            "on_LogWindowToolbarButtonClear_clicked"    : self.clear
        })
        
        self.content = self.get_widget("LogWindowContent")
        self.buttons = self.get_widget_dict(["Save", "Clear"], "LogWindowToolbarButton")
        self.filterboxes = self.get_widget_dict(["Level", "Source"], "LogWindowToolbarHBox")
        
        self.filters = {
            "level"     : self._createFilterBox("level", [_("Debug"), _("Info"), _("Warning"), _("Error")], 1),
            "source"    : self._createFilterBox("source", [_("LottaNZB and HellaNZB"), _("LottaNZB only"), _("HellaNZB only")], 1)
        }
        
        self.colormap = gtk.widget_get_default_colormap()
        
        errorTag = gtk.TextTag(name = "error")
        errorTag.set_property("foreground-gdk", self.colormap.alloc_color("red"))
        
        debugTag = gtk.TextTag(name = "debug")
        debugTag.set_property("foreground-gdk", self.colormap.alloc_color("grey"))
        
        self.tagTable = gtk.TextTagTable()
        self.tagTable.add(errorTag)
        self.tagTable.add(debugTag)
        
        self.buffer = gtk.TextBuffer(self.tagTable)
        self.content.set_buffer(self.buffer)
        
        self.show()
        
        self.handlerID = log.connect("changed", self.update)
        self.update()
    
    def destroy(self, *args):
        log.disconnect(self.handlerID)
        Window.destroy(self)
    
    def getRecords(self):
        source = self.getActiveSource()
        level = self.getActiveLevel()
        
        if source == LogWindow.SOURCE_LOTTA:
            store = log.getLottaStore()
        elif source == LogWindow.SOURCE_HELLA:
            store = log.getHellaStore()
        else:
            store = log
        
        store = store.filterByLevel(level)
        
        return store.records
    
    def getActiveSource(self):
        return self.filters["source"].get_active()
    
    def getActiveLevel(self):
        # Could be implemented more directly
        
        level = self.filters["level"].get_active()
        
        if level == LogWindow.LEVEL_DEBUG:
            return logging.DEBUG
        elif level == LogWindow.LEVEL_INFO:
            return logging.INFO
        elif level == LogWindow.LEVEL_WARNING:
            return logging.WARNING
        elif level == LogWindow.LEVEL_ERROR:
            return logging.ERROR
    
    def update(self, *args):
        if self.visible():
            self.buffer.set_text("")
            records = self.getRecords()
            
            for record in records:
                if self.getActiveSource() == LogWindow.SOURCE_BOTH:
                    if record.source == RECORD_SOURCE_LOTTA:
                        source = "LottaNZB\t"
                    else:
                        source = "HellaNZB\t"
                else:
                    source = ""
                
                line = strftime("%H:%M:%S", localtime(record.created)) + "\t" + source + record.msg + "\n"
                
                if record.levelno >= logging.WARNING:
                    tag = "error"
                elif record.levelno == logging.DEBUG:
                    tag = "debug"
                else:
                    tag = ""
                
                if tag:
                    self.buffer.insert_with_tags_by_name(self.buffer.get_end_iter(), line, tag)
                else:
                    self.buffer.insert(self.buffer.get_end_iter(), line)
    
    def save(self, *args):
        def handleResponse(dialog, response):
            if response == gtk.RESPONSE_OK:
                filename = dialog.get_filename()
                
                try:
                    logfile = open(filename, "w")
                except IOError, e:
                    logging.error(_("Could not open %s to write message log: %s") % (filename, e.strerror))
                    return
                
                try:
                    logfile.write(self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter()))
                    logfile.close()
                except IOError, e:
                    logging.error(_("Could not write message log to %s: %s") % (filename, e.strerror))
                else:
                    logging.info(_("Message log saved to %s.") % (filename))
            
            dialog.destroy()
        
        dialog = gtk.FileChooserDialog(_("Save message log"),
            self.window, gtk.FILE_CHOOSER_ACTION_SAVE,
            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        
        dialog.connect("response", handleResponse)
        dialog.set_default_response(gtk.RESPONSE_OK)
        dialog.set_alternative_button_order((gtk.RESPONSE_OK, gtk.RESPONSE_CANCEL))
        
        dialog.show()
    
    def clear(self, *args):
        log.clear()
    
    def _createFilterBox(self, name, entries, active = 0):
        filterBox = gtk.combo_box_new_text()
        
        for entry in entries:
            filterBox.append_text(entry)
        
        filterBox.set_active(active)
        filterBox.connect("changed", self.update)
        filterBox.show()
        
        self.filterboxes[name].pack_end(filterBox)
        
        return filterBox
    
    SOURCE_BOTH = 0
    SOURCE_LOTTA = 1
    SOURCE_HELLA = 2
    
    LEVEL_DEBUG = 0
    LEVEL_INFO = 1
    LEVEL_WARNING = 2
    LEVEL_ERROR = 3
