#!/usr/bin/env python

import sys, atexit
from optparse import OptionParser

import locale, gettext

locale.setlocale(locale.LC_ALL, "")

gettext.textdomain("lottanzb")
gettext.bindtextdomain("lottanzb")
gettext.install("lottanzb", unicode=1)

# Initialize logging functionality
import logging
from paths import clientDir

logging.basicConfig(level=logging.DEBUG, \
    format="%(asctime)s %(levelname)-8s %(message)s", \
    datefmt='%d.%m.%y %H:%M', \
    filename=clientDir("log"), \
    filemode="w")

# LottaNZB requires Python 2.4 or greater
try:
    assert float(sys.version[:3]) >= 2.4
except AssertionError:
    logging.error(_("You need to install Python 2.4 or greater to run LottaNZB."))
    sys.exit()

# Trying to import pyGTK
try:
    import pygtk
    pygtk.require("2.0")
    
    import gobject, gtk, gtk.glade
except:
    logging.error(_("You need to install pyGTK to run LottaNZB."))
    sys.exit()

gtk.glade.textdomain("lottanzb")
gtk.glade.bindtextdomain("lottanzb")

import modes

from __init__ import __version__
from gui import MainWindow
from util import Flock, ThreadManager
from backend import Backend

__lock = Flock("/tmp/lottanzb.lock")

class Lotta(ThreadManager):
    def __init__(self):
        logging.info(_("Starting LottaNZB %s using Python %s.") % (__version__, sys.version[:5]))
        
        ThreadManager.__init__(self)
        atexit.register(self._handleExit)
        
        self.registerThread(Backend())
        
        try:
            self.registerThread(modes.loadMode())
        except:
            pass
        
        self.leaveModeOnShutdown = True
    
    def showGUI(self):
        logging.debug(_("Launching GUI..."))
        
        MainWindow()
        
        gtk.main()
    
    def _handleExit(self):
        modes.saveActiveMode()
        
        # Not as elegant as it's supposed to be...
        if self.leaveModeOnShutdown:
            self.registerThread(modes.ModeActionThread()).join()
        
        self.stopAllThreads(True)
        
        logging.debug(_("Exiting..."))
        
        if gtk.main_level():
            gtk.main_quit()

def main():
    parser = OptionParser(usage = "%prog [FILES and/or NEWZBINIDS...]", version="LottaNZB %s" % __version__)
    parser.add_option("-d", "--debug", action="store_true", dest="debug", help="show debug messages", default=False)
    
    options, args = parser.parse_args()
    
    # Initialize console logging
    if options.debug:
        console_log_level = logging.DEBUG
    else:
        console_log_level = logging.INFO
    
    console = logging.StreamHandler()
    console.setLevel(console_log_level)
    console.setFormatter(logging.Formatter("%(levelname)-8s %(message)s"))
    
    logging.getLogger("").addHandler(console)
    
    # Temporary fix: prefs.py sometimes calls logging.error or logging.info while it's imported.
    # If the console logging isn't already set up at this point of time, we get a strange console output.
    from log import LottaHandler
    logging.getLogger("").addHandler(LottaHandler())
    
    # Check if another instance of LottaNZB is already running
    lockAcquired = __lock.acquire()
    
    files = []
    nzbIDs = []
    
    for arg in args:
        try:
            nzbIDs.append(int(arg))
        except ValueError:
            files.append(arg)
    
    commandLineAction = bool(len(files) or len(nzbIDs))
    
    if commandLineAction or lockAcquired: # If both are False, we've got nothing to do
        app = Lotta()
    
    if commandLineAction:
        def enqueueItems(*args):
            for file in files:
                try:
                    Backend().enqueue(file)
                except:
                    pass # An error message is already printed out by Backend().enqueue.
            
            for id in nzbIDs:
                try:
                    Backend().newzbinEnqueue(id)
                except:
                    pass # An error message is already printed out by Backend().newzbinEnqueue.
            
            if not lockAcquired:
                # Just to make sure the HellaNZB daemon isn't shut down by this (invisible and short-lived)
                # instance of LottaNZB.
                app.leaveModeOnShutdown = False
                sys.exit()
        
        if modes.getActiveMode():
            enqueueItems()
        else:
            modes.events.connect_temporarily("mode-changed", enqueueItems)
    
    if lockAcquired:
        app.showGUI()
    elif not commandLineAction:
        logging.error(_("LottaNZB is already running on this computer."))
        sys.exit()

if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
