import logging, atexit, gtk, os, sys, subprocess, time, copy, gobject
from inspect import isclass
import util, prefs, paths
from backend import Backend
from kiwigobject import PropertyObject, gsignal, type_register

_activeMode = None

class ModeActionThread(util.Thread):
    '''Entering and leaving usage modes takes place within this util.Thread, so that
    the UI isn't blocked. Pass a subclass of Mode respectively an instance of
    such a class to the constructor of this class or None, if you only want to
    leave the current usage mode.'''
    
    def __init__(self, mode=None):
        util.Thread.__init__(self)
        
        '''As we can't handle exceptions directly within the thread, they are
        stored in this property.'''
        self.exception = None
        
        '''May be either a subclass of Mode or an instance of such a class.'''
        self.mode = mode
    
    def run(self):
        global _activeMode
        
        try:
            modeInstance = None
            
            if self.mode:
                if isclass(self.mode):
                    modeInstance = self.mode()
                else:
                    modeInstance = self.mode
            
            oldMode = getActiveMode()
            
            if oldMode:
                try:
                    oldMode.leave(bool(modeInstance))
                except:
                    logging.error(_("Could not leave current usage mode '%s'.") % (oldMode))
                    raise
            
            _activeMode = None
            
            if modeInstance:
                try:
                    modeInstance.enter()
                except:
                    logging.error(_("Could not enter usage mode '%s'.") % (modeInstance))
                    raise
        
        except Exception, e:
            self.exception = e
        else:
            _activeMode = modeInstance
            
            if oldMode:
                if not modeInstance:
                    logging.info(_("Left usage mode '%s'.") % (oldMode))
                elif oldMode.__class__ is modeInstance.__class__:
                    logging.info(_("Reentered usage mode '%s'.") % (modeInstance))
                else:
                    logging.info(_("Changed usage mode from '%s' to '%s'.") % (oldMode, modeInstance))
            elif modeInstance:
                logging.info(_("Set usage mode to '%s'.") % (modeInstance))
        
            events.emit("mode-changed")
        
        self.emit("completed")

class __ModeEvents(gobject.GObject):
    __gsignals__ = { "mode-changed" : (gobject.SIGNAL_RUN_LAST, None, []) }
    
    def connect_temporarily(self, detailed_signal, handler, *args):
        handlerID = None
        
        def auto_disconnect(*args, **kwargs):
            self.disconnect(handlerID)
            handler(*args, **kwargs)
        
        handlerID = self.connect(detailed_signal, auto_disconnect, *args)

events = __ModeEvents()

def getActiveMode():
    return _activeMode

def getActiveModeClass():
    mode = getActiveMode()
    
    if mode:
        return mode.__class__

def saveActiveMode():
    mode = getActiveMode()
    
    if mode:
        logging.debug(_("Saving preferences..."))
        
        mode.save()

def loadMode():
    savedModeClass = getSavedModeClass()
    
    try:
        savedMode = savedModeClass()
    except TypeError:
        # Todo: Better error message
        logging.info(_("The LottaNZB configuration file couldn't be found or might be invalid. Launching usage mode selection window..."))
        showSelectionWindow(quitOnCancel=True, onCancelWarning=_("No usage mode has been selected. Shutting down LottaNZB..."))
    except Exception, e:
        showSelectionWindow(errorMessage=str(e), quitOnCancel=True)
    else:
        def handleCompleted(thread, *args):
            if thread.exception:
                showSelectionWindow(errorMessage=str(thread.exception), quitOnCancel=True)
        
        thread = ModeActionThread(savedMode)
        thread.connect("completed", handleCompleted)
        
        return thread

def getSavedModeClass():
    if prefs.lotta.loaded:
        if prefs.lotta["frontend_mode"]:
            if prefs.lotta["remote"]:
                return RemoteFrontendMode
            else:
                return LocalFrontendMode
        else:
            return StandaloneMode

def showSelectionWindow(errorMessage = "", quitOnCancel = False, onCancelWarning = ""):
    if errorMessage:
        logging.error(errorMessage)
    
    dialog = SelectionDialog(errorMessage)
    
    if quitOnCancel:
        def handleClosed(*args):
            if not getActiveMode():
                if onCancelWarning:
                    logging.warning(onCancelWarning)
                
                sys.exit()
        
        dialog.widget.connect("destroy", handleClosed)
    
    dialog.show()

class Mode:
    def __init__(self, **kwargs):
        self._eventmap = {
            "unexpectedDisconnect" : self.handleUnexpectedDisconnect
        }
        
        if "eventmap" in kwargs:
            self._eventmap.update(kwargs["eventmap"])
        
        self._handlerRef = {}
        
        if "hellaconfig" in kwargs:
            hellaPrefsLocation = kwargs["hellaconfig"]
        else:
            hellaPrefsLocation = None
        
        self.hellaPrefs = prefs.HellaPrefs(hellaPrefsLocation)
    
    def enter(self):
        for event, handler in self._eventmap.iteritems():
            self._handlerRef[event] = Backend().connect(event, handler)
        
        # We might drop this in the future
        prefs.hella = self.hellaPrefs
    
    def leave(self, *args):
        for event in self._eventmap.keys():
            Backend().disconnect(self._handlerRef[event])
        
        prefs.hella = prefs.HellaPrefs() # dummy config
    
    def save(self):
        prefs.lotta.save()
    
    def __str__(self):
        return self.__class__.name
    
    def handleUnexpectedDisconnect(self, *args):
        '''Needs to be implemented separately by sub classes.'''
        raise NotImplementedError
    
    def needsConfiguration(self):
        '''Needs to be implemented separately by sub classes.'''
        raise NotImplementedError
        
    @staticmethod
    def isActive():
        return isinstance(getActiveMode(), Mode)

class StandaloneMode(Mode):
    name = _("Stand-alone mode")
    short = _("Stand-alone")
    description = _("Use LottaNZB as a normal NZB Usenet client. When launching LottaNZB, HellaNZB is started in the background. You can easily change the HellaNZB preferences from within LottaNZB.")
    icon = gtk.STOCK_PREFERENCES
    
    def __init__(self):
        self.launcherCommand = prefs.lotta["hellanzb_launcher"]
        
        Mode.__init__(self, hellaconfig = paths.clientDir("hellanzb.conf"))
    
    def enter(self):
        assert self.launcherCommand and self.hellaPrefs.loaded and self.hellaPrefs.hasServers()
        
        # The HellaNZB preferences need to be written to the disk, so that HellaNZB can read them.
        # May be it's a little inefficient.
        self.hellaPrefs.save()
        
        self.startHella()
        Backend().connectToHella(*self.getConnectionArgs())
        
        Mode.enter(self)
    
    def getConnectionArgs(self):
        return (self.hellaPrefs.XMLRPC_SERVER, self.hellaPrefs.XMLRPC_PORT, self.hellaPrefs.XMLRPC_PASSWORD)
    
    def leave(self, cleanLeaving):
        Backend().disconnectFromHella()
        self.stopHella(cleanLeaving)
        
        Mode.leave(self)
    
    def handleUnexpectedDisconnect(self, *args):
        try:
            self.restartHella()
            Backend().connectToHella(*self.getConnectionArgs())
        except Exception, e:
            util.logException()
            showSelectionWindow(errorMessage=_("The connection to the HellaNZB daemon was lost unexpecedly and trying to restart and reconnect failed.") + "\n\n" + str(e), quitOnCancel=True)
    
    def save(self):
        self.hellaPrefs.save()
        
        prefs.lotta.update({
            "frontend_mode"     : False,
            "remote"            : False,
            "hellanzb_launcher" : self.launcherCommand
        })
        
        Mode.save(self)
    
    def usesCorrectConfigFile(self):    
        try:
            proxy = Backend().checkConnection(*self.getConnectionArgs())
            lines = proxy.status()["log_entries"]
            
            # The XMLRPC method status() doesn't return more than 20 log entries
            # that's why the log entry we're interested in will be missing in
            # some cases. Parsing the log file is the only solution.
            if len(lines) == 20:
                logFile = self.hellaPrefs.LOG_FILE
                log = open(logFile, "r").read()
                
                # Todo: This is still incomplete
                i1 = log.rindex("INFO hellanzb v")
                i2 = log.rindex("LOGFILE")
                lines = [{ "INFO": log[i1:i2] }]
            
            for type, line in lines[0].iteritems():
                if line.index(self.hellaPrefs.confFile):
                    return True
        except:
            pass
        
        return False
    
    def startHella(self):
        generalMessage = _("Could not start the HellaNZB daemon.")
        
        try:
            self.fixMissingSSLModule()
            
            if self.usesCorrectConfigFile():
                logging.info(_("The HellaNZB daemon is already running using the configuration file managed by LottaNZB."))
                return
            
            process = subprocess.Popen(self.launcherCommand + " -D -c " + self.hellaPrefs.confFile, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            
            if process.wait():
                line = process.stderr.readline()
                
                if line:
                    if line.find("Address already in use") != -1:
                        raise AlreadyRunningError, _("The port %s is already in use. Another HellaNZB daemon might be already running.") % (self.hellaPrefs.XMLRPC_PORT)
                    else:
                        raise Exception, line
                else:
                    raise Exception, _("The HellaNZB daemon could not be started for an unknown reason.")
            
            logging.info(_("Started HellaNZB daemon."))
        except (OSError, ValueError), e:
            logging.error(generalMessage)
            
            if e.errno == 2:
                raise Exception, _("Could not find the HellaNZB executable (%s).") % (self.launcherCommand)
            else:
                raise Exception, str(e)
        except:
            logging.error(generalMessage)
            raise
    
    def stopHella(self, waitForCompleteShutdown = False):
        try:
            # The output of the HellaNZB executable is passed to grep, which will
            # return anything in this case. We don't want to show the output to the user.
            subprocess.Popen(self.launcherCommand + " -c " + self.hellaPrefs.confFile + " shutdown | grep \"nothingatall\"", shell=True)
            
            # Try to avoid starting an instance of HellaNZB while an old one is still being shut down.
            if waitForCompleteShutdown:
                running = True
                
                while running:
                    try:
                        # This will raise an exception as soon as the connection
                        # can't be established anymore. It's not a very elegant
                        # approach but it's certainly better than the old one.
                        Backend().checkConnection(*self.getConnectionArgs())
                    except:
                        running = False
                    
                    time.sleep(0.2)
        
        # This shouldn't happen in general.
        except (OSError, ValueError), e:
            logging.error(_("Could not stop the HellaNZB daemon."))
            raise
        else:
            self.process = None
            logging.info(_("HellaNZB daemon stopped."))
    
    # Some magic behind the scenes
    def fixMissingSSLModule(self):
        try:
            import OpenSSL
        except ImportError:
            for server in self.hellaPrefs.servers:
                server.ssl = False
    
    def restartHella(self):
        self.stopHella(True)
        self.startHella()
        
        logging.info(_("HellaNZB daemon restarted."))
    
    @staticmethod
    def isActive():
        return isinstance(getActiveMode(), StandaloneMode)

class AlreadyRunningError(Exception):
    pass

class FrontendMode(Mode):
    def __init__(self, *args, **kwargs):
        Mode.__init__(self, *args, **kwargs)
        self.hellaPrefs.read_only = True
    
    def enter(self):
        Backend().connectToHella(*self.getConnectionArgs())
        Mode.enter(self)
    
    def handleUnexpectedDisconnect(self, *args, **kwargs):
        try:
            Backend().connectToHella(*self.getConnectionArgs())
        except Exception, e:
            showSelectionWindow(errorMessage=_("The connection to the HellaNZB daemon was lost unexpecedly and the attempt to reconnect failed.") + " " + kwargs["hint"] + "\n\n" + str(e), quitOnCancel=True)
    
    def leave(self, *args):
        Backend().disconnectFromHella()
        Mode.leave(self)
    
    def save(self):
        prefs.lotta["frontend_mode"] = True
        
        Mode.save(self)
        
    @staticmethod
    def isActive():
        return isinstance(getActiveMode(), FrontendMode)

class LocalFrontendMode(FrontendMode):
    name = _("Local front-end mode")
    short = _("Local front-end")
    description = _("Monitor a HellaNZB daemon on this computer, which is already configured and running.")
    icon = gtk.STOCK_CONNECT
    
    def __init__(self):
        FrontendMode.__init__(self, hellaconfig = prefs.lotta["frontend_config_file"])
    
    def enter(self):
        assert self.hellaPrefs.loaded
        
        FrontendMode.enter(self)
    
    def getConnectionArgs(self):
        return [self.hellaPrefs.XMLRPC_SERVER, self.hellaPrefs.XMLRPC_PORT, self.hellaPrefs.XMLRPC_PASSWORD]
    
    def save(self):
        prefs.lotta["remote"] = False
        FrontendMode.save(self)
    
    def handleUnexpectedDisconnect(self, *args):
        FrontendMode.handleUnexpectedDisconnect(self, hint = _("Please make sure HellaNZB is running and the correct configuration file is selected."))
    
    @staticmethod
    def isActive():
        return isinstance(getActiveMode(), LocalFrontendMode)

class RemoteFrontendMode(FrontendMode):
    name = _("Remote front-end mode")
    short = _("Remote front-end")
    description = _("Monitor a HellaNZB daemon on another computer, which is already configured and running.")
    icon = gtk.STOCK_NETWORK
    
    def __init__(self):
        # Cache...
        self.connectionArgs = [prefs.lotta["xmlrpc_address"], prefs.lotta["xmlrpc_port"], prefs.lotta["xmlrpc_password"]]
        
        FrontendMode.__init__(self)
    
    def enter(self):
        args = self.getConnectionArgs()
        assert args[0] and args[1] and args[2]
        
        FrontendMode.enter(self)
    
    def save(self):
        args = self.getConnectionArgs()
        
        prefs.lotta.update({
            "remote"            : False,
            "xmlrpc_address"    : args[0],
            "xmlrpc_port"       : args[1],
            "xmlrpc_password"   : args[2]
        })
        
        FrontendMode.save(self)
    
    def getConnectionArgs(self):
        return self.connectionArgs
    
    def handleUnexpectedDisconnect(self, *args):
        FrontendMode.handleUnexpectedDisconnect(self, hint = _("Please make sure HellaNZB is running and the connection information are correctly set."))
    
    @staticmethod
    def isActive():
        return isinstance(getActiveMode(), RemoteFrontendMode)

class SelectionDialog(util.Window):
    def __init__(self, error = ""):
        self.initialError = error
        
        util.Window.__init__(self, "ModesDialog", {
            "on_ModesDialog_response"                               : self.handleResponse,
            "on_ModesDialogModeList_cursor_changed"                 : self.handleModeChange,
            "on_ModesDialogStandaloneAuthentication_toggled"        : self.handleAuthToggle,
            "on_ModesDialogFrontendLocalConfiguration_file_set"     : self.validateCurrentMode
        })
        
        # The current active usage mode. This will enable LottaNZB to restore it the user
        # hits 'Cancel'.
        self.oldMode = getActiveMode()
        
        # Create an instance of each usage mode.
        self.modes = [StandaloneMode, LocalFrontendMode, RemoteFrontendMode]
        
        self.modeTree = self.get_widget("ModesDialogModeList")
        self.buttons = self.get_widget_dict(["Cancel", "Save"], "ModesDialogButton")
        self.description = self.get_widget_dict(["Header", "Label"], "ModesDialogDescription")
        
        self.error = self.get_widget_dict(["Header", "Label"], "ModesDialogError")
        self.errorWidget = self.get_widget("ModesDialogError")
        
        self.hint = self.get_widget("ModesDialogHint")
        self.selectionInfoPane = self.get_widget("ModesDialogSelection")
        self.specific = self.get_widget_dict(["Standalone", "FrontendRemote", "FrontendLocal"], "ModesDialog")
        
        self.fields = {
            StandaloneMode      : self.get_widget_dict(["Address", "Port", "Authentication", "Username", "Password", "Executable"], "ModesDialogStandalone"),
            LocalFrontendMode   : self.get_widget_dict(["Configuration"], "ModesDialogFrontendLocal"),
            RemoteFrontendMode  : self.get_widget_dict(["Address", "Port", "Password"], "ModesDialogFrontendRemote")
        }
        
        labels = self.get_widget_dict(["AddressLabel", "UsernameLabel", "ExecutableLabel"], "ModesDialogStandalone")
        map(gtk.SizeGroup(gtk.SIZE_GROUP_HORIZONTAL).add_widget, labels.values())
        
        currentModeClass = getActiveModeClass() or getSavedModeClass()
        
        if prefs.lotta.loaded:
            self.lottaPrefsBackup = copy.deepcopy(prefs.lotta)
        else:
            self.setDefaultLottaPrefs()
        
        self.lastSelectedMode = None
        
        self.createModeList()
        self.handleAuthToggle()
        self.fillForms()
        self.setupFormValidation()
        
        if currentModeClass:
            self.selectMode(currentModeClass)
        
        self.showUsageHint(not bool(currentModeClass))
    
    def createModeList(self):
        self.modeList = gtk.ListStore(str, str)
        self.modeTree.set_model(self.modeList)
        
        for mode in self.modes:
            self.modeList.append([mode.icon, mode.short])
        
        column = gtk.TreeViewColumn(_("Usage modes"))
        iconCell = gtk.CellRendererPixbuf()
        nameCell = gtk.CellRendererText()
        
        column.pack_start(iconCell, False)
        column.pack_start(nameCell, True)
        
        column.set_attributes(iconCell, stock_id = 0)
        column.set_attributes(nameCell, text = 1)
        
        self.modeTree.append_column(column)
        self.modeTree.set_search_column(1)
    
    def handleAuthToggle(self, *args):
        requiresAuth = self.fields[StandaloneMode]["authentication"].get_active()
        
        for fieldName in ["username", "password"]:
            self.fields[StandaloneMode][fieldName].set_sensitive(requiresAuth)
            
            if not requiresAuth:
                self.fields[StandaloneMode][fieldName].set_text("")
    
    def getSelectedMode(self):
        model, selection_iter = self.modeTree.get_selection().get_selected()
        
        if selection_iter:
            return self.modes[model.get_path(selection_iter)[0]]
    
    def selectMode(self, mode):
        if mode:
            self.modeTree.get_selection().select_path((self.modes.index(mode), ))
    
    def fillForms(self):
        standaloneMode = self.createModeInstance(StandaloneMode)
        
        if standaloneMode.hellaPrefs.hasServers():
            server = standaloneMode.hellaPrefs.servers[0]
            host = server.getDefaultHost()
            
            self.fields[StandaloneMode]["authentication"].set_active(server.needsAuthentication())
            
            self.fields[StandaloneMode]["address"].set_text(host.address)
            self.fields[StandaloneMode]["port"].set_value(int(host.port))
            
            if server.needsAuthentication():
                self.fields[StandaloneMode]["username"].set_text(server.username)
                self.fields[StandaloneMode]["password"].set_text(server.password)
        
        self.fields[StandaloneMode]["executable"].set_text(prefs.lotta["hellanzb_launcher"] or "")
        
        if not prefs.lotta["xmlrpc_address"] == "localhost":
            self.fields[RemoteFrontendMode]["address"].set_text(prefs.lotta["xmlrpc_address"] or "")
            self.fields[RemoteFrontendMode]["port"].set_value(int(prefs.lotta["xmlrpc_port"]))
            self.fields[RemoteFrontendMode]["password"].set_text(prefs.lotta["xmlrpc_password"] or "")
        
        confFile = prefs.lotta["frontend_config_file"]
        
        if confFile:
            self.fields[LocalFrontendMode]["configuration"].set_filename(confFile)
    
    def setupFormValidation(self):
        self.validators = {}
        
        for mode, fields in self.fields.iteritems():
            validator = util.LottaForm(fields)
            
            def handler(form, valid, mode):
                self.handleValidation(mode, form.valid)
            
            validator.connect("notify::valid", handler, mode)
            
            self.validators[mode] = validator
    
    def validateCurrentMode(self, *args):
        mode = self.getSelectedMode()
        
        if mode:
            valid = self.validators[mode].validate()
        else:
            valid = False
        
        self.handleValidation(mode, valid)
    
    def handleValidation(self, mode, valid):
        if mode is self.getSelectedMode():
            self.buttons["save"].set_sensitive(valid)
    
    def handleModeChange(self, *args):
        mode = self.getSelectedMode()
        
        if mode is self.lastSelectedMode:
            return
        
        if mode:
            self.selectionInfoPane.show()
            
            self.description["header"].set_markup("<b>" + mode.name + "</b>")
            self.description["label"].set_markup(mode.description)
            
            for name, widget in self.specific.iteritems():
                widget.hide()
            
            if mode is StandaloneMode:
                self.specific["standalone"].show()
            elif mode is LocalFrontendMode:
                self.specific["frontendlocal"].show()
            else:
                self.specific["frontendremote"].show()
        else:
            self.selectionInfoPane.hide()
        
        if self.initialError:
            self.showError(self.initialError)
            self.initialError = ""
        else:
            self.hideError()
        
        self.lastSelectedMode = mode
        self.validateCurrentMode()
    
    def handleResponse(self, widget, response):
        if response == gtk.RESPONSE_APPLY:
            self.handleSave()
        else:
            self.restoreAndClose()
    
    def createModeInstance(self, ModeClass):
        mode = ModeClass()
        
        if ModeClass is StandaloneMode:
            if not mode.hellaPrefs.loaded:
                # Todo: Need better implementation
                read_only = mode.hellaPrefs.read_only
                confFile = mode.hellaPrefs.confFile
                
                mode.hellaPrefs = prefs.HellaPrefs(prefs.HellaPrefs.locate())
                
                mode.hellaPrefs.read_only = read_only
                mode.hellaPrefs.confFile = confFile
                mode.hellaPrefs.loaded = True
            
            mode.hellaPrefs.SKIP_UNRAR = not bool(mode.hellaPrefs.UNRAR_CMD)
            
            if mode.hellaPrefs.hasServers():
                server = mode.hellaPrefs.servers[0]
                
                if server.id == "changeme" and server.username == "changeme":
                    mode.hellaPrefs.servers.remove(server)
        else:
            if ModeClass is LocalFrontendMode:
                pass
            else:
                pass
        
        return mode
    
    def handleSave(self):
        self.disableButtons()
        
        try:
            ModeClass = self.getSelectedMode()
            
            if ModeClass is StandaloneMode:
                prefs.lotta["hellanzb_launcher"] = self.fields[StandaloneMode]["executable"].get_text()
            else:
                if ModeClass is LocalFrontendMode:
                    prefs.lotta["frontend_config_file"] = self.fields[LocalFrontendMode]["configuration"].get_filename()
                else:
                    prefs.lotta["xmlrpc_address"] = self.fields[RemoteFrontendMode]["address"].get_text()
                    prefs.lotta["xmlrpc_port"] = int(self.fields[RemoteFrontendMode]["port"].get_value())
                    prefs.lotta["xmlrpc_password"] = self.fields[RemoteFrontendMode]["password"].get_text()
            
            mode = self.createModeInstance(ModeClass)
            
            if ModeClass is StandaloneMode:
                serverInfo = {
                    "username" : self.fields[StandaloneMode]["username"].get_text(),
                    "password" : self.fields[StandaloneMode]["password"].get_text(),
                    "hosts"    : [self.fields[StandaloneMode]["address"].get_text() + ":" + str(int(self.fields[StandaloneMode]["port"].get_value()))],
                }
                
                if mode.hellaPrefs.hasServers():
                    server = mode.hellaPrefs.servers[0]
                    server.update(serverInfo)
                else:
                    serverInfo["id"] = _("Default Server")
                    mode.hellaPrefs.addServer(serverInfo)
            
            thread = ModeActionThread(mode)
            thread.start()
            thread.join()
            
            if thread.exception:
                raise thread.exception
        
        except Exception, e:
            self.showError(str(e))
            logging.error(str(e))
        else:
            saveActiveMode()
            self.close()
        
        self.enableButtons()
        
    def showError(self, message, header = None):
        if header is None:
            header = _("Error")
        
        self.errorWidget.show()
        self.error["header"].set_markup("<b>" + header + "</b>")
        self.error["label"].set_text(message.strip())
    
    def hideError(self):
        self.errorWidget.hide()
    
    def disableButtons(self):
        for button in self.buttons.values():
            button.set_sensitive(False)
        
        gtk.main_iteration()
    
    def enableButtons(self):
        for button in self.buttons.values():
            button.set_sensitive(True)
        
        gtk.main_iteration()
    
    def restoreAndClose(self, *args):
        try:
            prefs.lotta.update(self.lottaPrefsBackup)
        except:
            pass
        
        if not self.initialError and self.oldMode and getActiveMode() is not self.oldMode:
            # We don't need to wait for the thread to complete. It's better to close the window immediately.
            ModeActionThread(self.oldMode).start()
        
        self.close()
    
    def close(self):
        self.destroy()
    
    def showUsageHint(self, autoChange = False):
        from backend import WrongPasswordError
        
        try:
            Backend().checkConnection("localhost", 8760, "")
        except WrongPasswordError:
            running = True
        except:
            running = False
        else:
            running = True
        
        if running:
            if self.createModeInstance(StandaloneMode).usesCorrectConfigFile():
                hint = _("LottaNZB is currently in stand-alone mode. HellaNZB is already running and uses the configuration file managed by LottaNZB.")
                mode = StandaloneMode
            else:
                hint = _("The HellaNZB daemon seems to be already running. The local front-end mode might be the most appropriate usage mode for you.")
                mode = LocalFrontendMode
        else:
            if self.getHellaCommand():
                hint = _("No HellaNZB daemon seems to be listening on the default port, so it's most likely not running. The stand-alone mode might be the most appropriate usage mode for you.")
                mode = StandaloneMode
            else:
                hint = _("The HellaNZB executable could not be found. This could mean that it isn't installed on your system or that you've installed it in a non-standard way. If you want to control a HellaNZB daemon on another computer, select the remote front-end usage mode. Otherwise, make sure HellaNZB is installed properly.")
                mode = RemoteFrontendMode
        
        if autoChange:
            self.selectMode(mode)
        
        self.hint.set_text(hint)
    
    def getHellaCommand(self):
        # Cache the value to avoid unnecessary disk accesses
        if hasattr(self, "_hellaCommand"):
            return self._hellaCommand
        
        places = ["/usr/bin/hellanzb", "/usr/bin/hellanzb.py", "/usr/local/bin/hellanzb", "/usr/local/bin/hellanzb.py"]
        
        for place in places:
            if os.path.isfile(place):
                self._hellaCommand = place
                return self._hellaCommand
        
        self._hellaCommand = None
    
    def setDefaultLottaPrefs(self):
        prefs.lotta.update({
            "sleep_time"                : 1.0,
            "window_height"             : 350,
            "window_width"              : 700,
            "start_minimized"           : False,
            "hellanzb_launcher"         : self.getHellaCommand(),
            "xmlrpc_address"            : "localhost",
            "xmlrpc_port"               : 8760,
            "xmlrpc_password"           : "changeme",
            "prefs_revision"            : 1
        })
        
        prefs.lotta.loaded = True
