import gtk, os, logging
import util, prefs, modes
from backend import Backend

class StandaloneModeDialog:
    def __init__(self):
        self.serverList = gtk.ListStore(str, str, str)
        self.oldHellaPrefs = prefs.HellaPrefs(prefs.hella.confFile)
        
        self.uiTree = util.UITree("Preferences")
        self.window = self.uiTree.get_root_widget()
        
        self.uiTree.signal_autoconnect({
            "on_Preferences_response"                       : self.handleResponse,
            "on_PreferencesGeneralMaxRateCheckbox_toggled"  : self.handleMaxRateToggle,
            "on_PreferencesServerButtonEdit_clicked"        : self.editServer,
            "on_PreferencesServerButtonAdd_clicked"         : self.addServer,
            "on_PreferencesServerButtonRemove_clicked"      : self.removeServer,
            "on_PreferencesServerTreeView_cursor_changed"   : self.updateServerButtons,
            "on_PreferencesServerTreeView_row_activated"    : self.editServer,
            "on_PreferencesNewzbinEnabled_toggled"          : self.handleNewzbinToggle
        })
        
        self.generalFields = self.uiTree.get_widget_dict(["DownloadDir", "MaxRateCheckbox", "MaxRate", "MaxRateScale", "SmartPar", "Notify", "Minimized"], "PreferencesGeneral")
        self.newzbinFields = self.uiTree.get_widget_dict(["Enabled", "Username", "Password"], "PreferencesNewzbin")
        self.newzbinAuthWidgets = self.uiTree.get_widget_dict(["Username", "UsernameLabel", "Password", "PasswordLabel"], "PreferencesNewzbin")
        self.serverButtons = self.uiTree.get_widget_dict(["Add", "Edit", "Remove"], "PreferencesServerButton")
        
        self.serverTreeView = self.uiTree.get_widget("PreferencesServerTreeView")
        self.serverTreeView.set_model(self.serverList)
        self.serverTreeView.set_rules_hint(True)
        self.serverTreeView.set_reorderable(True)
        self.serverTreeView.show()
        
        for id, title in enumerate([_("Name"), _("Address"), _("Username")]):
            column = gtk.TreeViewColumn(title, gtk.CellRendererText(), text=id)
            column.set_resizable(True)
            column.set_expand(True)
            self.serverTreeView.append_column(column)
        
        self.fill()
        self.window.run()
    
    def hide(self):
        self.window.hide()
    
    def fill(self):
        self.generalFields["downloaddir"].set_current_folder(prefs.hella.DEST_DIR)
        
        maxRate = prefs.hella.MAX_RATE
        
        if maxRate:
            self.generalFields["maxratecheckbox"].set_active(True)
            self.generalFields["maxrate"].set_value(maxRate)
        else:
            self.generalFields["maxratecheckbox"].set_active(False)
        
        checkboxes = {
            "smartpar"  : prefs.hella.SMART_PAR,
            "notify"    : prefs.hella.LIBNOTIFY_NOTIFY,
            "minimized" : prefs.lotta["start_minimized"]
        }
        
        for field, preference in checkboxes.iteritems():
            if preference:
                active = True
            else:
                active = False
            
            self.generalFields[field].set_active(active)
        
        # Now we fill in the values on the Newzbin.com tab
        # Fill in the enable checkbox and the username and password
        newzbinEnabled = prefs.hella.hasNewzBinSupport()
        
        self.newzbinFields["enabled"].set_active(newzbinEnabled)
        self.handleNewzbinToggle()
        
        if newzbinEnabled:
            self.newzbinFields["username"].set_text(prefs.hella.NEWZBIN_USERNAME)
            self.newzbinFields["password"].set_text(prefs.hella.NEWZBIN_PASSWORD)
        
        self.fillServerList()
    
    def fillServerList(self):
        selectedServer = self.getSelectedServer()
        
        self.serverList.clear()
        
        for server in prefs.hella.servers:
            self.serverList.append([server.id, server.getDefaultHost().address, server.username])
        
        self.selectServer(selectedServer)
        self.updateServerButtons()
    
    def handleResponse(self, widget, response):
        if response == gtk.RESPONSE_APPLY:
            self.save()
        else:
            self.cancel()
    
    def save(self, *args):
        if self.generalFields["maxratecheckbox"].get_active():
            maxRate = int(self.generalFields["maxrate"].get_value() or 0)
        else:
            maxRate = 0
        
        maxRateChanged = maxRate != prefs.hella.MAX_RATE
        
        prefs.hella.update({
            "MAX_RATE"         : maxRate,
            "DEST_DIR"         : self.generalFields["downloaddir"].get_current_folder(),
            "SMART_PAR"        : self.generalFields["smartpar"].get_active(),
            "LIBNOTIFY_NOTIFY" : self.generalFields["notify"].get_active(),
            "NEWZBIN_USERNAME" : self.newzbinFields["username"].get_text(),
            "NEWZBIN_PASSWORD" : self.newzbinFields["password"].get_text()
        })
        
        prefs.lotta["start_minimized"] = self.generalFields["minimized"].get_active()
        prefs.lotta["newzbin"] = self.newzbinFields["enabled"].get_active()
        prefs.lotta.save()
        
        self.oldHellaPrefs.MAX_RATE = prefs.hella.MAX_RATE # This can be changed using XMLRPC
        
        if self.oldHellaPrefs != prefs.hella:
            prefs.hella.save()
            modes.ModeActionThread(modes.StandaloneMode).start()
        elif maxRateChanged:
            Backend().setMaxRate(maxRate)
        
        self.hide()
    
    def cancel(self, *args):
        prefs.hella = self.oldHellaPrefs
        
        self.hide()
    
    def handleMaxRateToggle(self, widget):
        active = widget.get_active()
        
        self.generalFields["maxrate"].set_property("sensitive", active)
        self.generalFields["maxratescale"].set_property("sensitive", active)
    
    def handleNewzbinToggle(self, *args):
        active = self.newzbinFields["enabled"].get_active()
        
        for field, widget in self.newzbinAuthWidgets.iteritems():
            widget.set_property("sensitive", active)
        
        if not active:
            self.newzbinFields["username"].set_text("")
            self.newzbinFields["password"].set_text("")
    
    def updateServerButtons(self, *args):
        server = self.getSelectedServer()
        
        for field in ["edit", "remove"]:
            self.serverButtons[field].set_property("sensitive", server)
    
    def getSelectedServer(self):
        model, selection_iter = self.serverTreeView.get_selection().get_selected()
        
        try:
            return prefs.hella.servers[model.get_path(selection_iter)[0]]
        except:
            pass
    
    def selectServer(self, server):
        try:
            self.serverTreeView.get_selection().select_path((prefs.hella.servers.index(server), ))
        except:
            pass
    
    def editServer(self, *args):
        server = self.getSelectedServer()
        
        if server:
            dialog = EditServerDialog(server)
            self.fillServerList()
    
    def addServer(self, *args):
        dialog = AddServerDialog()
        self.fillServerList()
    
    def removeServer(self, *args):
        server = self.getSelectedServer()
        
        if server and util.Message.question(_("Are you sure you want to delete this server?")):
            try:
                prefs.hella.servers.remove(server)
                prefs.hella.save()
                
                self.fillServerList()
            except:
                logging.error(_("Could not delete server."))

class AbstractServerDialog:
    def __init__(self):
        self.uiTree = util.UITree("ServerDialog")
        self.window = self.uiTree.get_root_widget()
        
        self.fields = self.uiTree.get_widget_dict(["Name", "Address", "Port", "AuthRequired", "Username", "Password", "SSL", "Connections"])
        self.buttons = self.uiTree.get_widget_dict(["Cancel", "Save"], "ServerDialogButton")
        self.labels = self.uiTree.get_widget_dict(["NameLabel", "UsernameLabel"], "ServerDialog")
        
        self.uiTree.signal_autoconnect({
            "on_ServerDialog_response"              : self.handleResponse,
            "on_ServerDialogAuthRequired_toggled"   : self.handleAuthToggle,
            "on_ServerDialogSSL_toggled"            : self.handleSSLToggle,
        })
        
        map(gtk.SizeGroup(gtk.SIZE_GROUP_HORIZONTAL).add_widget, self.labels.values())
        
        def handleValidationChange(form, *args):
            self.buttons["save"].set_sensitive(form.valid)
        
        import re
        
        def checkServerName(input):
            input = re.compile('[^A-Za-z0-9\-_() ]').sub("", input)
            inputOK = False
            
            if input:
                inputOK = True
            
            return (inputOK, input)
        
        form = util.LottaForm(self.fields, { "name": checkServerName })
        form.connect("notify::valid", handleValidationChange)
        
        form.validate()
    
    def handleAuthToggle(self, *args):
        requiresAuth = self.fields["authrequired"].get_active()
        
        for fieldName in ["username", "password"]:
            self.fields[fieldName].set_sensitive(requiresAuth)
            
            if not requiresAuth:
                self.fields[fieldName].set_text("")
    
    def handleSSLToggle(self, *args):
        if self.fields["ssl"].get_active():
            try:
                import OpenSSL
            except ImportError:
                util.Message.error(_("OpenSSL not available"), _("In order to use SSL, you need to install the Python wrapper around OpenSSL. On Ubuntu, the package name is 'python-openssl'."))
                self.fields["ssl"].set_active(False)
    
    def getInput(self):
        input = {
            "id"          : self.fields["name"].get_text(),
            "hosts"       : [self.fields["address"].get_text() + ":" + str(int(self.fields["port"].get_value()))],
            "username"    : self.fields["username"].get_text(),
            "password"    : self.fields["password"].get_text(),
            "ssl"         : self.fields["ssl"].get_active(),
            "connections" : self.fields["connections"].get_value()
        }
        
        return input
    
    def close(self):
        self.window.destroy()
    
    def handleResponse(self, widget, response):
        if response == gtk.RESPONSE_APPLY:
            self.saveHandler()
        else:
            self.close()
    
    def saveHandler(self):
        self.save()
    
    def save(self):
        pass

class EditServerDialog(AbstractServerDialog):
    def __init__(self, server):
        AbstractServerDialog.__init__(self)
        
        self.server = server
        self.fill()
        
        self.window.set_title(_("Edit server")) 
        self.window.run()
    
    def fill(self):
        host = self.server.getDefaultHost()
        
        self.fields["name"].set_text(self.server.id)
        self.fields["address"].set_text(host.address)
        self.fields["port"].set_value(int(host.port))
        
        self.fields["ssl"].set_active(self.server.ssl)
        self.fields["connections"].set_value(self.server.connections)
        
        if self.server.needsAuthentication():
            self.fields["authrequired"].set_active(True)
            self.fields["username"].set_text(self.server.username)
            self.fields["password"].set_text(self.server.password)
        else:
            self.fields["authrequired"].set_active(False)
    
    def save(self):
        self.server.update(self.getInput())
        self.close()

class AddServerDialog(AbstractServerDialog):
    def __init__(self):
        AbstractServerDialog.__init__(self)
        
        self.window.set_title(_("Add a new server"))
        self.window.run()
    
    def save(self):
        prefs.hella.addServer(self.getInput())
        self.close()
