import os, sys, struct, traceback, gtk, gtk.glade, logging, re, gobject, threading
import paths
from kiwigobject import PropertyObject, gsignal, gproperty, type_register

gobject.threads_init()

# http://www.johnstowers.co.nz/blog/index.php/2007/03/12/threading-and-pygtk/
# (c) 2008, John Stowers <john.stowers@gmail.com>
class IdleObject(PropertyObject, gobject.GObject):
    """
    Override gobject.GObject to always emit signals in the main thread
    by emitting on an idle handler
    """
    
    __gtype_name__ = "IdleObject"
    
    def __init__(self):
        gobject.GObject.__init__(self)
        PropertyObject.__init__(self)
    
    def emit(self, *args):
        gobject.idle_add(gobject.GObject.emit, self, *args)

type_register(IdleObject)

class Thread(threading.Thread, IdleObject):
    """
    Cancellable thread which uses gobject signals to return information
    to the GUI.
    """
    
    __gtype_name__ = "Thread"
    
    gsignal("completed", flags=gobject.SIGNAL_RUN_LAST)
    gsignal("progress", retval=float, flags=gobject.SIGNAL_RUN_LAST) # Percent complete
    
    def __init__(self):
        threading.Thread.__init__(self)
        IdleObject.__init__(self)
        
        self.threadStopped = threading.Event()
    
    def stop(self):
        """
        Threads in python are not cancellable, so we implement our own
        cancellation logic
        """
        self.threadStopped.set()
    
    def run(self):
        raise NotImplementedError
        
        # Example code:
        # while not self.threadStopped.isSet():
        #   ...
        #   self.emit("progress", x)
        # self.emit("completed")

type_register(Thread)

class ThreadManager:
    def __init__(self):
        # Stores all threads
        self.threads = []
    
    def _handleThreadCompleted(self, thread, *args):
        self.threads.remove(thread)
    
    def registerThread(self, thread, **kwargs):
        # Signals run in the order connected. Connect the user completed
        # callback first in case they wish to do something 
        # before we delete the thread
        if "onComplete" in kwargs:
            thread.connect("completed", kwargs["onComplete"], None)
        
        thread.connect("completed", self._handleThreadCompleted, None)
        
        if "onProgress" in kwargs:
            thread.connect("progress", kwargs["onProgress"], None)
        
        self.threads.append(thread)
        
        if not thread.isAlive():
            thread.start()
        
        return thread
    
    def stopAllThreads(self, block=False):
        """
        Stops all threads. If block is True then actually wait for the thread
        to finish (may block the UI) 
        """
        for thread in self.threads:
            thread.stop()
            
            if block and thread.isAlive():
                try:
                    thread.join()
                except:
                    pass

# Most efficient generic function for currying
# From http://www.daniweb.com/code/snippet564.html
def curry(fn, *cargs, **ckwargs):
    def call_fn(*fargs, **fkwargs):
        d = ckwargs.copy()
        d.update(fkwargs)
        return fn(*(cargs + fargs), **d)
    
    return call_fn

class UITree(gtk.glade.XML):
    def __init__(self, root):
        self.root = root
        
        gtk.glade.XML.__init__(self, paths.gladeDir("lottanzb.glade"), root)
    
    def get_widget_dict(self, widgets, root=""):
        dict = {}
        
        for widget in widgets:
            dict[widget.lower()] = self.get_widget((root or self.root) + widget)
        
        return dict
    
    def get_root_widget(self):
        return self.get_widget(self.root)

class WidgetHelper(UITree):
    def __init__(self, root, signals = None):
        UITree.__init__(self, root)
        
        self.widget = self.get_root_widget()
        self.toggling = False
        
        if signals:
            self.signal_autoconnect(signals)

    def hide(self, *args):
        self.widget.hide()
    
    def destroy(self, *args):
        self.widget.destroy()
    
    def show(self, *args):
        self.widget.show()
    
    def toggle(self, *args):
        if self.toggling:
            return False
        
        self.toggling = True
        
        if self.visible():
            self.hide()
        else:
            self.show()
            
        self.toggling = False
        
        return True
    
    def visible(self):
        return self.window.get_property("visible")

class Window(WidgetHelper):
    def __init__(self, *args):
        WidgetHelper.__init__(self, *args)
        
        self.window = self.widget

class Message(gtk.MessageDialog):
    def __init__(self, message, secondary, **args):
        if "buttons" in args:
            buttons = getattr(gtk, "BUTTONS_" + args["buttons"].upper())
        else:
            buttons = gtk.BUTTONS_OK
        
        if "type" in args:
            type = getattr(gtk, "MESSAGE_" + args["type"].upper())
        else:
            type = gtk.MESSAGE_INFO
        
        gtk.MessageDialog.__init__(self, message_format=message, parent=None, \
            flags=gtk.DIALOG_MODAL, buttons=buttons, type=type)
        
        if secondary:
            self.format_secondary_markup(secondary)
        
        self.response = self.run()
        self.destroy()
    
    @staticmethod
    def info(message, secondary = ""):
        Message(message, secondary)
    
    @staticmethod
    def question(message, secondary = ""):
        return Message(message, secondary, type="question", buttons="ok_cancel").response == gtk.RESPONSE_OK
    
    @staticmethod
    def warning(message, secondary = ""):
        Message(message, secondary, type="warning")
    
    @staticmethod
    def error(message, secondary = ""):
        Message(message, secondary, type="error")

# From http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/107747
from UserDict import UserDict

class OrderedDict(UserDict):
    def __init__(self, dict = None):
        self._keys = []
        UserDict.__init__(self, dict)
    
    def __delitem__(self, key):
        UserDict.__delitem__(self, key)
        self._keys.remove(key)
    
    def __setitem__(self, key, item):
        UserDict.__setitem__(self, key, item)
        if key not in self._keys: self._keys.append(key)
    
    def clear(self):
        UserDict.clear(self)
        self._keys = []
    
    def copy(self):
        dict = UserDict.copy(self)
        dict._keys = self._keys[:]
        return dict
    
    def items(self):
        return zip(self._keys, self.values())
    
    def keys(self):
        return self._keys[:]
    
    def reorder(self, keys):
        for oldkey in self._keys:
            if not oldkey in keys:
                del self[oldkey]
        
        self._keys = keys
    
    def popitem(self):
        try:
            key = self._keys[-1]
        except IndexError:
            raise KeyError("dictionary is empty")
        
        val = self[key]
        del self[key]
        
        return (key, val)
    
    def setdefault(self, key, failobj = None):
        UserDict.setdefault(self, key, failobj)
        if key not in self._keys: self._keys.append(key)
    
    def update(self, dict):
        UserDict.update(self, dict)
        
        for key in dict.keys():
            if key not in self._keys: self._keys.append(key)
    
    def values(self):
        return map(self.get, self._keys)
    
    def len(self):
        return len(self._keys)
    
    def getIndexById(self, id):
        try:
            return self._keys.index(id)
        except IndexError:
            pass
    
    def index(self, item):
        try:
            return self.values().index(item)
        except IndexError:
            pass
    
    def getByIndex(self, index):
        return self[self.getIdByIndex(index)]
    
    def getIdByIndex(self, index):
        return self._keys[index]
    
    def first(self):
        return self.getByIndex(0)
    
    def last(self):
        return self.getByIndex(self.len() - 1)
    
    def has(self, value):
        return self.index(value) is not None

def __checkerPrototype(regex, input):
    input = re.compile(regex).sub("", input)
    inputOK = False
    
    if input:
        inputOK = True
    
    return (inputOK, input)

LottaFormCheckers = {
    "numerical"    : curry(__checkerPrototype, "[^0-9]"),
    "onlyascii"    : curry(__checkerPrototype, "[^A-Za-z0-9\-_() ]")
}

class LottaForm(PropertyObject, gobject.GObject):
    gproperty("valid", bool, True)
    gproperty("validating", bool, False)
    
    def __init__(self, fields, checkers = None):
        gobject.GObject.__init__(self)
        PropertyObject.__init__(self)
        
        self._fields = fields
        self._checkers = checkers or {}
        
        for field, widget in self._fields.iteritems():
            if hasattr(widget, "get_text"):
                widget.connect("changed", self.validate)
            elif hasattr(widget, "get_active"):
                widget.connect("toggled", self.validate)
    
    def validate(self, *args):
        if self.validating:
            return
        
        self.validating = True
        
        inputOK = True
        
        for field, widget in self._fields.iteritems():
            valueMethod = ""
            
            for method in ["get_text", "get_filename"]:
                if hasattr(widget, method):
                    valueMethod = method
                    break
            
            if valueMethod and widget.get_property("sensitive"):
                input = getattr(widget, valueMethod)()
                
                if not input:
                    inputOK = False
                elif field in self._checkers:
                    valid, newInput = self._checkers[field](input)
                    
                    if valid == False:
                        inputOK = False
                    
                    if valueMethod == "get_text" and newInput != input:
                        widget.set_text(newInput)
        
        if self.valid != inputOK:
            self.valid = inputOK
        
        self.validating = False
        
        return self.valid

def logException(maxTBlevel = 3):
     cla, exc, trbk = sys.exc_info()
     excName = cla.__name__
     
     excTb = traceback.format_tb(trbk, maxTBlevel)
     
     try:
        logging.error(excName + ": " + exc[0])
     except:
        pass
     
     for excTbLine in excTb:
        logging.debug(excTbLine.strip())

# From http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/498171
from socket import gethostname

class Flock(object):
    '''Class to handle creating and removing (pid) lockfiles'''
    
    # custom exceptions
    class FileLockAcquisitionError(Exception): pass
    class FileLockReleaseError(Exception): pass
    
    # convenience callables for formatting
    addr = lambda self: "%d@%s" % (self.pid, self.host)
    fddr = lambda self: "<%s %s>" % (self.path, self.addr())
    pddr = lambda self, lock: "<%s %s@%s>" % (self.path, lock["pid"], lock["host"])
    
    def __init__(self, path, debug=None):
        self.pid = os.getpid()
        self.host = gethostname()
        self.path = path
        self.debug = debug # set this to get status messages
    
    def acquire(self):
        '''Acquire a lock, returning self if successful, False otherwise'''
        if self.islocked():
            if self.debug:
                lock = self._readlock()
                print "Previous lock detected: %s" % self.pddr(lock)
            
            return False
        try:
            fh = open(self.path, "w")
            fh.write(self.addr())
            fh.close()
            
            if self.debug:
                print "Acquired lock: %s" % self.fddr()
        except:
            if os.path.isfile(self.path):
                try:
                    os.unlink(self.path)
                except:
                    pass
            raise (self.FileLockAcquisitionError, "Error acquiring lock: %s" % self.fddr())
        return self
    
    def release(self):
        '''Release lock, returning self'''
        if self.ownlock():
            try:
                os.unlink(self.path)
                if self.debug:
                    print "Released lock: %s" % self.fddr()
            except:
                raise (self.FileLockReleaseError, "Error releasing lock: %s" % self.fddr())
        
        return self
    
    def _readlock(self):
        '''Internal method to read lock info'''
        try:
            lock = {}
            fh = open(self.path)
            data = fh.read().rstrip().split("@")
            fh.close()
            lock["pid"], lock["host"] = data
            return lock
        except:
            return { "pid": 8**10, "host": "" }
    
    def islocked(self):
        '''Check if we already have a lock'''
        try:
            lock = self._readlock()
            os.kill(int(lock["pid"]), 0)
            return (lock["host"] == self.host)
        except:
            return False
    
    def ownlock(self):
        '''Check if we own the lock'''
        lock = self._readlock()
        return (self.fddr() == self.pddr(lock))
    
    def __del__(self):
        '''Magic method to clean up lock when program exits'''
        self.release()

