#!/usr/bin/env python
from distutils.core import setup, Command
from distutils.dir_util import remove_tree
from distutils.command.install import install as _install
from distutils.command.build import build as _build
from distutils.command.clean import clean as _clean
from distutils.command.bdist_rpm import bdist_rpm as _bdist_rpm
from distutils.errors import *
from subprocess import call, Popen, PIPE
from glob import glob
from os.path import isfile, isdir, join, basename, exists, abspath
import os, msgfmt

from imp import load_module, find_module
lottanzb = load_module("lottanzb", *find_module("lottanzb"))

NAME             = "lottanzb"
VERSION          = lottanzb.__version__
DESCRIPTION      = "LottaNZB - Automated Usenet Client"
LONG_DESCRIPTION = "LottaNZB is a Usenet client that automates the download of Usenet files with the help of NZB files. It uses HellaNZB as its backend and PyGTK for its user interface."
AUTHOR           = "LottaNZB Development Team"
AUTHOR_EMAIL     = "avirulence@lottanzb.org"
URL              = "http://www.lottanzb.org/"
LICENSE          = "GPL"
KEYWORDS         = ["usenet", "nzb", "newzbin", "download", "hellanzb", "frontend"]
CLASSIFIERS      = [
                        "Development Status :: 4 - Beta",
                        "Environment :: X11 Applications :: Gnome",
                        "Environment :: X11 Applications :: GTK",
                        "Intended Audience :: End Users/Desktop",
                        "License :: OSI Approved :: GNU General Public License (GPL)",
                        "Operating System :: POSIX :: Linux",
                        "Natural Language :: Dutch",
                        "Natural Language :: English",
                        "Natural Language :: French",
                        "Natural Language :: German",
                        "Natural Language :: Spanish",
                        "Programming Language :: Python",
                        "Topic :: Communications :: Usenet News"
                   ]
DATA_FILES       = [
                        ("share/lottanzb/glade", glob("glade/*.*")),
                        ("share/doc/lottanzb", ["CHANGELOG", "LICENSE"])
                   ]

class CleanCommand(_clean):
    description = "remove the build and dist dir as well as pyc files"
    
    def run(self):
        _clean.run(self)
        
        def remove(files):
            for file in files:
                try:
                    os.unlink(file)
                except:
                    pass
        
        for root, dirs, files in os.walk("."):
            remove([join(root, file) for file in files if file.endswith(".pyc")])
        
        remove(["MANIFEST", ".xml2po.mo"])
        
        if self.all and exists(self.build_base):
            remove_tree(self.build_base)

class InstallWrapper(_install):
    def completeInstallation(self):
        def do(cmd, msg):
            try:
                print msg
                call(cmd)
            except:
                self.warn("Could not call %s." % cmd)
        
        do(["update-mime-database", join(self.prefix, "share", "mime")], "Updating MIME types...")
        do(["update-desktop-database"], "Updating desktop database...")
        
        for loc in ["gnome", "hicolor"]:
            do(["gtk-update-icon-cache", "-q", "-f", "-t", join(self.prefix, "share", "icons", loc)], "Updating icon caches...")
        
        do(["scrollkeeper-update", "-q", "-o", join(self.prefix, "share", "omf")], "Update scrollkeeper database...")

INSTALLED_FILES = "INSTALLED_FILES"

class InstallCommand(InstallWrapper):
    def initialize_options(self):
        InstallWrapper.initialize_options(self)
        
        self.record = None
    
    def finalize_options(self):
        InstallWrapper.finalize_options(self)
        
        if not self.record:
            self.record = INSTALLED_FILES
    
    def run(self):
        _install.run(self)
        self.completeInstallation()
        
        if self.record and isfile(self.record):
            print "\n\nLottaNZB can be uninstalled using 'python setup.py uninstall'."
            print "This command depends on the newly created file %s." % self.record

class UpgradeCommand(InstallCommand):
    def run(self):
        def try_to_call(cmd):
            try:
                call(cmd)
            except:
                pass
        
        map(try_to_call, [
            ["rm", "-f", join(self.prefix, "bin", "lottanzb")],
            ["rm", "-f", join(self.prefix, "share", "pixmaps", "lottanzb.png")]
        ])
        
        process = Popen(["find", join(self.prefix, "lib"), "-wholename", join(self.prefix, "lib", "python2.*", "site-packages", "lottanzb-0.*-py2.*.egg")], stdout=PIPE)
        
        if not process.wait():
            map(try_to_call, [["rm", "-rf", dir.strip()] for dir in process.stdout.readlines()])
        
        InstallCommand.run(self)

class UninstallCommand(InstallWrapper):
    def run(self):
        try:
            files = file(INSTALLED_FILES, "r").readlines()
        except:
            raise DistutilsFileError("Could not find list of installed files (%s)." % INSTALLED_FILES)
        
        for f in [f.strip() for f in files]:
            try:
                os.unlink(f)
            except OSError, e:
                if e.errno == 13:
                    raise DistutilsFileError("Root privileges are required to uninstall LottaNZB.")
                elif e.errno == 2:
                    self.warn("Could not find file %s." % f)
            else:
                self.announce("Removing %s..." % f)
        
        os.remove(INSTALLED_FILES)
        self.completeInstallation()
        
        print "LottaNZB has been uninstalled successfully."

class BuildExtra(_build):
    """Adds the extra commands to the build target."""
    
    def finalize_options(self):
        _build.finalize_options(self)
        
        self.sub_commands.append(("build_i18n",  None))
        self.sub_commands.append(("build_icons", None))
        self.sub_commands.append(("build_help",  None))

class BuildIcons(Command):
    description = "select all icons for installation"
    user_options = [("icon-dir=", "i", "icon directory of the source tree")]
    
    def initialize_options(self):
        self.icon_dir = None
    
    def finalize_options(self):
        if self.icon_dir is None:
            self.icon_dir = join("data", "icons")
    
    def run(self):
        def add(theme, size, category, pattern):
            self.distribution.data_files.append(( \
                join("share", "icons", theme, size, category), \
                glob(join(self.icon_dir, size, category, pattern))))
        
        for size_path in glob(join(self.icon_dir, "*")):
            size = basename(size_path)
            
            for category_path in glob(join(size_path, "*")):
                category = basename(category_path)
                pattern = "*.png"
                
                if category == "apps":
                    add("hicolor", size, category, pattern)
                elif category == "mimetypes":
                    if size == "scalable":
                        pattern = "*.*"
                    
                    add("gnome", size, category, pattern)

class BuildI18N(Command):
    description = "integrate the gettext framework"
    boolean_options = ["merge-po"]
    user_options = [("desktop-files=",   None, ".desktop.in files that should be merged"),
                    ("xml-files=",       None, ".xml.in files that should be merged"),
                    ("schemas-files=",   None, ".schemas.in files that should be merged"),
                    ("ba-files=",        None, "bonobo-activation files that should be merged"),
                    ("rfc822deb-files=", None, "RFC822 files that should be merged"),
                    ("key-files=",       None, ".key.in files that should be merged"),
                    ("domain=",          "d",  "gettext domain"),
                    ("merge-po",         "m",  "merge po files against template"),
                    ("po-dir=",          "p",  "directory that holds the i18n files")]
    
    def initialize_options(self):
        self.desktop_files = []
        self.xml_files = []
        self.key_files = []
        self.schemas_files = []
        self.ba_files = []
        self.rfc822deb_files = []
        self.domain = None
        self.merge_po = False
        self.po_dir = None
    
    def finalize_options(self):
        if self.domain is None:
            self.domain = self.distribution.metadata.name
        if self.po_dir is None:
            self.po_dir = "po"
    
    def run(self):
        """Update the language files, generate mo files and add them
           to the to be installed files"""
        
        data_files = self.distribution.data_files
        
        # Update po(t) files and print a report
        # We have to change the working dir to the po dir for intltool
        if self.has_intltool():
            cmd = ["intltool-update", (self.merge_po and "-r" or "-p"), "-g", self.domain]
            wd = os.getcwd()
            os.chdir(self.po_dir)
            call(cmd)
            os.chdir(wd)
        
        for po_file in glob(join(self.po_dir, "*.po")):
            lang = basename(po_file[:-3])
            mo_dir = join("build", "mo", lang, "LC_MESSAGES")
            mo_file = join(mo_dir, "%s.mo" % self.domain)
            
            if not exists(mo_dir):
                os.makedirs(mo_dir)
            
            msgfmt.make(po_file, mo_file)
            
            targetpath = join("share", "locale", lang, "LC_MESSAGES")
            data_files.append((targetpath, [mo_file]))
        
        # Merge .in with translation
        for (option, switch) in ((self.xml_files, "-x"),
                                 (self.desktop_files, "-d"),
                                 (self.schemas_files, "-s"),
                                 (self.rfc822deb_files, "-r"),
                                 (self.ba_files, "-b"),
                                 (self.key_files, "-k"),):
            try:
                file_set = eval(option)
            except:
                continue
            for (target, files) in file_set:
                build_target = join("data", target)
                
                if not exists(build_target): 
                    os.makedirs(build_target)
                
                files_merged = []
                
                for file in files:
                    if file.endswith(".in"):
                        file_merged = basename(file[:-3])
                    else:
                        file_merged = basename(file)
                    
                    file_merged = join(build_target, file_merged)
                    
                    if self.has_intltool():
                        call(["intltool-merge", switch, self.po_dir, file, file_merged])
                    
                    files_merged.append(file_merged)
                
                data_files.append((target, files_merged))
    
    @staticmethod
    def has_intltool():
        try:
            return call(["which", "intltool-merge"]) == 0
        except:
            return False

class BuildHelp(Command):
    description = "install a docbook based documentation"
    user_options = [("help-dir=", None, "help directory of the source tree"),
                    ("update",   "u",   "update po and pot files and generate translated xml files")]
    
    boolean_options = ["update"]
    
    def initialize_options(self):
        self.help_dir = None
        self.update = False
    
    def finalize_options(self):
        if self.help_dir is None:
            self.help_dir = "help"
    
    def run(self):
        def help_dir(*args):
            return join(self.help_dir, *args)
        
        langs = [basename(path) for path in glob(help_dir("*")) if isdir(path) and not basename(path) == "po"]
        
        original_omf = help_dir("C", "lottanzb-C.omf")
        original_xml = help_dir("C", "lottanzb.xml")
        pot_file = help_dir("po", "documentation.pot")
        
        if self.update:
            if self.has_xml2po():
                call(["xml2po", "-k", "-o", pot_file, original_xml, original_omf])
                
                for lang in langs:
                    if not lang == "C":
                        dest_omf = abspath(help_dir(lang, "lottanzb-%s.omf" % lang))
                        dest_xml = help_dir(lang, "lottanzb.xml")
                        dest_po = help_dir("po", "%s.po" % lang)
                        
                        call(["msgmerge", "-U", dest_po, pot_file])
                        call(["xml2po", "-k", "-p", dest_po, "-o", dest_xml, original_xml])
                        call(["xml2po", "-k", "-p", dest_po, "-o", dest_omf, original_omf])
                        
                        process = Popen("sed 's/\"C\"/\"%s\"/' %s | sed 's/\/C\//\/%s\//'" % (lang, dest_omf, lang), shell=True, stdout=PIPE)
                        
                        if not process.wait():
                            file = open(dest_omf, "w")
                            file.write(process.stdout.read())
            else:
                self.error("Please install the necessary gnome-doc-utils package first.")
        
        data_files = self.distribution.data_files
        app_name = self.distribution.metadata.name
        
        for lang in langs:
            path_xml = join("share", "gnome", "help", app_name, lang)
            path_figures = join("share", "gnome", "help", app_name, lang, "figures")
            
            data_files.append((path_xml, (glob(help_dir(lang, "*.xml")))))
            data_files.append((path_figures, (glob(help_dir(lang, "figures", "*.png")))))
        
        data_files.append((join("share", "omf", app_name), glob(help_dir("*", "*.omf"))))
    
    @staticmethod
    def has_xml2po():
        try:
            return call(["which", "xml2po"]) == 0
        except:
            return False

class BuildRPM(_bdist_rpm):
    def finalize_options(self):
        _bdist_rpm.finalize_options(self)
        
        if isfile("/etc/mandriva-release"):
            dist_release = open("/etc/mandriva-release").read().split(" ")[3]
            
            self.distribution_name = "Mandriva Linux %s" % dist_release
            self.release = "%s.mdv%s" % (self.release, dist_release)
            self.requires = "python >= 2.4 hellanzb >= 0.13 pygtk2.0 pygtk2.0-libglade"
        
        elif isfile("/etc/fedora-release"):
            dist_release = open("/etc/fedora-release").read().split(" ")[2]
            
            self.distribution_name = "Fedora %s" % dist_release
            self.release = "%s.fc%s" % (self.release, dist_release)
            self.requires = "python >= 2.4 hellanzb >= 0.13 pygtk2 pygtk2-libglade"
setup(
      name             = NAME,
      version          = VERSION,
      description      = DESCRIPTION,
      long_description = LONG_DESCRIPTION,
      author           = AUTHOR,
      author_email     = AUTHOR_EMAIL,
      url              = URL,
      packages         = ["lottanzb"],
      license          = LICENSE,
      scripts          = ["data/lottanzb"],
      keywords         = KEYWORDS,
      data_files       = DATA_FILES,
      classifiers      = CLASSIFIERS,
      cmdclass         = {
                            "build"         : BuildExtra,
                            "build_i18n"    : BuildI18N,
                            "build_help"    : BuildHelp,
                            "build_icons"   : BuildIcons,
                            "install"       : InstallCommand,
                            "uninstall"     : UninstallCommand,
                            "clean"         : CleanCommand,
                            "upgrade"       : UpgradeCommand,
                            "bdist_rpm"     : BuildRPM
                         }
)
